# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import csv
import logging
import sys
from collections import OrderedDict
from collections.abc import Iterable

import click
import yaml
import yaml.resolver

from datacube.cfg import ODCEnvironment
from datacube.index import Index
from datacube.index.abstract import AbstractIndex
from datacube.ui import click as ui
from datacube.ui.click import cli
from datacube.utils import gen_password
from datacube.utils.serialise import SafeDatacubeDumper

_LOG: logging.Logger = logging.getLogger("datacube-user")
USER_ROLES = ("user", "manage", "admin")


@cli.group(name="user", help="User management commands")
def user_cmd() -> None:
    pass


def build_user_list(index):
    lstdct = []
    for role, user, description in index.users.list_users():
        info = OrderedDict(
            (("role", role), ("user", user), ("description", description))
        )
        lstdct.append(info)
    return lstdct


def _write_csv(index: Iterable) -> None:
    writer = csv.DictWriter(
        sys.stdout, ["role", "user", "description"], extrasaction="ignore"
    )
    writer.writeheader()

    def add_first_role(row):
        roles_ = row["role"]
        row["role"] = roles_ if roles_ else None
        return row

    writer.writerows(add_first_role(row) for row in index)


def _write_yaml(index: Iterable) -> None:
    """
    Dump yaml data with support for OrderedDicts.

    Allows for better human-readability of output: such as dataset ID field first, sources last.

    (Ordered dicts are output identically to normal yaml dicts: their order is purely for readability)
    """
    return yaml.dump_all(
        index, sys.stdout, SafeDatacubeDumper, default_flow_style=False, indent=4
    )


_OUTPUT_WRITERS = {
    "csv": _write_csv,
    "yaml": _write_yaml,
}


@user_cmd.command("list")
@click.option(
    "-f",
    help="Output format",
    type=click.Choice(list(_OUTPUT_WRITERS)),
    default="yaml",
    show_default=True,
)
@ui.pass_index()
def list_users(index: Index, f: str) -> None:
    """
    List users
    """
    _OUTPUT_WRITERS[f](build_user_list(index))


@user_cmd.command("grant")
@click.argument("role", type=click.Choice(USER_ROLES), nargs=1)
@click.argument("users", nargs=-1)
@ui.pass_index()
def grant(index: Index, role: str, users: Iterable[str]) -> None:
    """
    Grant a role to users
    """
    index.users.grant_role(role, *users)


@user_cmd.command("create")
@click.argument("role", type=click.Choice(USER_ROLES), nargs=1)
@click.argument("user", nargs=1)
@click.option("--description")
@ui.pass_index()
@ui.pass_config
def create_user(
    cfg_env: ODCEnvironment,
    index: AbstractIndex,
    role: str,
    user: str,
    description: str,
) -> None:
    """
    Create a User
    """
    password = gen_password(12)
    index.users.create_user(user, password, role, description=description)

    click.echo(
        "{host}:{port}:*:{username}:{password}".format(
            host=index.url_parts.hostname or "localhost",
            port=index.url_parts.port,
            username=user,
            password=password,
        )
    )


@user_cmd.command("delete")
@click.argument("users", nargs=-1)
@ui.pass_index()
@ui.pass_config
def delete_user(config, index: Index, users: str) -> None:
    """
    Delete a User
    """
    index.users.delete_user(*users)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import inspect
import math
from collections.abc import Callable, Generator, Sequence
from os import PathLike
from pathlib import Path
from types import SimpleNamespace
from typing import Any

import numpy as np
import toolz
import xarray as xr
from odc.geo import wh_
from odc.geo.geobox import GeoBox, zoom_to
from odc.geo.geobox import pad as gbox_pad
from odc.geo.warp import resampling_s2rio
from odc.geo.xr import xr_coords
from typing_extensions import override

from ..api import Datacube
from ..index.eo3 import EO3Grid, is_doc_eo3
from ..model import Dataset
from ..storage import BandInfo, reproject_and_fuse
from ..storage._read import rdr_geobox
from ..storage._rio import RasterDatasetDataSource, RasterioDataSource
from . import suppress_deprecations


class RasterFileDataSource(RasterioDataSource):
    """This is only used in test code"""

    def __init__(
        self, filename, bandnumber, nodata=None, crs=None, transform=None, lock=None
    ) -> None:
        super().__init__(filename, nodata, lock=lock)
        self.bandnumber = bandnumber
        self.crs = crs
        self.transform = transform

    @override
    def get_bandnumber(self, src):
        return self.bandnumber

    @override
    def get_transform(self, shape):
        if self.transform is None:
            raise RuntimeError("No transform in the data and no fallback")
        return self.transform

    @override
    def get_crs(self):
        if self.crs is None:
            raise RuntimeError("No CRS in the data and no fallback")
        return self.crs


def _raster_metadata(band):
    source = RasterDatasetDataSource(band)
    with source.open() as rdr:
        return SimpleNamespace(
            dtype=rdr.dtype.name, nodata=rdr.nodata, geobox=rdr_geobox(rdr)
        )


def get_raster_info(
    ds: Dataset, measurements: list[str] | None = None
) -> dict[str, Any]:
    """
    :param ds: Dataset
    :param measurements: List of band names to load
    """
    if measurements is None:
        measurements = list(ds.product.measurements)

    return {n: _raster_metadata(BandInfo(ds, n)) for n in measurements}


def eo3_geobox(ds: Dataset, band: str | None = None, grid: str = "default") -> GeoBox:
    if band is not None:
        mm = ds.measurements.get(ds.product.canonical_measurement(band), None)
        if mm is None:
            raise ValueError(f"No such band: {band}")
        grid = mm.get("grid", "default")

    crs = ds.crs
    doc_path = ("grids", grid)

    grid_spec = toolz.get_in(doc_path, ds.metadata_doc)
    if crs is None or grid_spec is None:
        raise ValueError("Not a valid EO3 dataset")

    parsed = EO3Grid(grid_spec)
    return GeoBox(parsed.shape, parsed.transform, crs)


def native_geobox(ds: Dataset, measurements=None, basis: str | None = None) -> GeoBox:
    """Compute native GeoBox for a set of bands for a given dataset

    :param ds: Dataset
    :param measurements: List of band names to consider
    :param basis: Name of the band to use for computing reference frame, other
    bands might be reprojected if they use different pixel grid

    :return: GeoBox describing native storage coordinates.
    """
    with suppress_deprecations():
        gs = ds.product.grid_spec
        if gs is not None:
            # Dataset is from ingested product, figure out GeoBox of the tile this dataset covers
            bb = [gbox for _, gbox in gs.tiles(ds.bounds)]
            if len(bb) != 1:
                # Ingested product but dataset overlaps several/none tiles -- no good
                raise ValueError("Broken GridSpec detected")
            return bb[0]

    if measurements is None and basis is None:
        measurements = list(ds.product.measurements)

    if is_doc_eo3(ds.metadata_doc):
        if basis is not None:
            return eo3_geobox(ds, basis)

        gboxes = [eo3_geobox(ds, band) for band in measurements]
    else:
        if basis is not None:
            return get_raster_info(ds, [basis])[basis].geobox

        ii = get_raster_info(ds, measurements)
        gboxes = [info.geobox for info in ii.values()]

    geobox = gboxes[0]
    consistent = all(geobox == gbox for gbox in gboxes)
    if not consistent:
        raise ValueError("Not all bands share the same pixel grid")
    return geobox


def compute_native_load_geobox(
    ds: Dataset,
    band: str,
    dst_geobox: GeoBox | None = None,
    buffer: float | None = None,
) -> GeoBox:
    """Compute area of interest for an input dataset given final output geobox.

    Take native projection and resolution from ``ds, band`` pair and compute
    region in that projection that fully encloses footprint of the
    ``dst_geobox`` with ``buffer``. Construct GeoBox that encloses that
    region fully with resolution/pixel alignment copied from supplied band.

    :param ds: Sample dataset (only resolution and projection is used, not footprint)
    :param band: Reference band to use
    :param dst_geobox:
                 (resolution of output GeoBox will match resolution of this band)
    :param buffer: Buffer in units of CRS of ``ds`` (meters usually) to cover enough aoi
        when ``dst_geobox`` is different than that of input ``ds``,
        default is 10 pixels worth
    """
    native = native_geobox(ds, basis=band)
    if dst_geobox is None:
        return native

    if buffer is None:
        buffer = 10 * max(map(abs, (native.resolution.y, native.resolution.x)))

    assert native.crs is not None
    return GeoBox.from_geopolygon(
        dst_geobox.extent.to_crs(native.crs).buffer(buffer),
        crs=native.crs,
        resolution=native.resolution,
        align=native.alignment,
    )


def _split_by_grid(xx: xr.DataArray) -> list[xr.DataArray]:
    """Split datasets by grid/crs"""

    def extract(grid_id, ii):
        yy = xx[ii]
        crs = xx.grid2crs[grid_id]
        yy.attrs.update(crs=crs)
        yy.attrs.pop("grid2crs", None)
        return yy

    return [extract(grid_id, ii) for grid_id, ii in xx.groupby(xx.grid).groups.items()]


def _native_load_1(
    sources: xr.DataArray,
    bands: tuple[str, ...],
    *,
    dst_geobox: GeoBox | None = None,
    optional_bands: tuple[str, ...] | None = None,
    basis: str | None = None,
    pad: int | None = None,
    **kw,
) -> xr.Dataset:
    """Load datasets with native crs
    :param sources: grouped datasets
    :param bands: List of band names to load
    :param dst_geobox: Geobox of final output, if None then use geobox of input dataset
    :param optional_bands: List of optional band names to load
    :param basis: Name of the band to use for computing reference frame, other
    bands might be reprojected if they use different pixel grid
    :param pad: number of pixels to pad the geobox adjusted to the requirement of reproject

    :param kw: Any other parameter that load_data accepts

    :return: Xarray dataset
    """
    if basis is None:
        basis = bands[0]
    (ds,) = sources.data[0]
    load_geobox = compute_native_load_geobox(ds, basis, dst_geobox)
    if pad is not None:
        load_geobox = gbox_pad(load_geobox, pad)

    mm = ds.product.lookup_measurements(bands)
    if optional_bands is not None:
        for ob in optional_bands:
            try:
                om = ds.product.lookup_measurements(ob)
            except KeyError:
                continue
            else:
                mm.update(om)

    xx = Datacube.load_data(sources, load_geobox, mm, **kw)
    return xx


def native_load(
    dss: Sequence[Dataset],
    bands: Sequence[str],
    groupby: Callable[..., Any],
    *args,
    dst_geobox: GeoBox | None = None,
    optional_bands: tuple[str, ...] | None = None,
    basis: str | None = None,
    pad: int | None = None,
    **kw,
) -> xr.Dataset | Generator[xr.Dataset, None, None]:
    """Load datasets in native resolution.

    :param dss: Datasets
    :param bands: List of band names to load
    :param groupby: Function to group the datasets
    :param args: positional passed into gropuby

    :param dst_geobox: Geobox of final output, if None then use geobox of input dataset
    :param optional_bands: List of optional band names to load
    :param basis: Name of the band to use for computing reference frame, other
    bands might be reprojected if they use different pixel grid
    :param pad: number of pixels to pad the geobox

    :param kw: Any other parameter groupby or _native_load_1 accepts

    :return: Xarray dataset or generator of Xarray dataset
    """
    # Filter **kw to match what groupby op accepts
    sig = inspect.signature(groupby)
    accepted_kw = {
        name
        for name, param in sig.parameters.items()
        if param.kind
        in (param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD, param.VAR_KEYWORD)
    }
    accepted_kw = {k: v for k, v in kw.items() if k in accepted_kw}
    sources = groupby(list(dss), *args, **accepted_kw)

    for key in accepted_kw:
        kw.pop(key, None)

    # split datasets if they are in different grid/crs
    if "grid" in sources.coords:

        def yield_by_grid():
            for srcs in _split_by_grid(sources):
                _xx = _native_load_1(
                    srcs,
                    tuple(bands),
                    dst_geobox=dst_geobox,
                    optional_bands=optional_bands,
                    basis=basis,
                    pad=pad,
                    **kw,
                )
                yield _xx

        return yield_by_grid()
    else:
        return _native_load_1(
            sources,
            tuple(bands),
            dst_geobox=dst_geobox,
            optional_bands=optional_bands,
            basis=basis,
            pad=pad,
            **kw,
        )


def dc_read(
    path,
    band: int = 1,
    geobox=None,
    resampling: str = "nearest",
    dtype=None,
    dst_nodata: float | None = None,
    fallback_nodata=None,
):
    """
    Use default io driver to read file without constructing Dataset object.
    """
    source = RasterFileDataSource(path, band, nodata=fallback_nodata)
    with source.open() as rdr:
        dtype = rdr.dtype if dtype is None else dtype
        if geobox is None:
            geobox = rdr_geobox(rdr)
        if dst_nodata is None:
            dst_nodata = rdr.nodata

    # currently dst_nodata = None case is not supported. So if fallback_nodata
    # was None and file had none set, then use 0 as default output fill value
    if dst_nodata is None:
        dst_nodata = 0

    im = np.full(geobox.shape, dst_nodata, dtype=dtype)
    reproject_and_fuse([source], im, geobox, dst_nodata, resampling=resampling)
    return im


def write_gtiff(
    fname: str | PathLike[str] | Path,
    pix: np.ndarray[tuple[int, ...], np.dtype[np.float64 | np.signedinteger[Any]]],
    crs: str = "epsg:3857",
    resolution=(10, -10),
    offset: tuple[float, float] = (0.0, 0.0),
    nodata=None,
    overwrite: bool = False,
    blocksize: int | None = None,
    geobox=None,
    **extra_rio_opts,
) -> SimpleNamespace:
    """Write ndarray to GeoTiff file.

    Geospatial info can be supplied either via
    - resolution, offset, crs
    or
    - geobox (takes precedence if supplied)
    """
    # pylint: disable=too-many-locals
    import rasterio
    from affine import Affine

    if pix.ndim == 2:
        h, w = pix.shape
        nbands = 1
        band: int | tuple[int, ...] = 1
    elif pix.ndim == 3:
        nbands, h, w = pix.shape
        band = tuple(i for i in range(1, nbands + 1))
    else:
        raise ValueError("Need 2d or 3d ndarray on input")

    if not isinstance(fname, Path):
        fname = Path(fname)

    if fname.exists():
        if overwrite:
            fname.unlink()
        else:
            raise OSError("File exists")

    if geobox is not None:
        assert geobox.shape == (h, w)

        A = geobox.transform
        crs = str(geobox.crs)
    else:
        sx, sy = resolution
        tx, ty = offset

        A = Affine(sx, 0, tx, 0, sy, ty)

    rio_opts = {
        "width": w,
        "height": h,
        "count": nbands,
        "dtype": pix.dtype.name,
        "crs": crs,
        "transform": A,
        "predictor": 2,
        "compress": "DEFLATE",
        # Rasterio 1.4.4 and earlier sets "INIT_DEST" to "NO_DATA". If there is no
        # NoData value set, GDAL 3.11 returns an error, and older versions assumed
        # NoData was 0. The next line sets nodata to NaN/0 to ensure all GDAL versions
        # behave the same way.
        "nodata": nodata
        if nodata is not None
        else math.nan
        if pix.dtype.kind == "f"
        else 0,
    }

    if blocksize is not None:
        rio_opts.update(
            tiled=True, blockxsize=min(blocksize, w), blockysize=min(blocksize, h)
        )

    rio_opts.update(extra_rio_opts)

    with rasterio.open(str(fname), "w", driver="GTiff", **rio_opts) as dst:
        dst.write(pix, band)
        meta = dst.meta

    meta["geobox"] = geobox if geobox is not None else rio_geobox(meta)
    meta["path"] = fname
    return SimpleNamespace(**meta)


def dc_crs_from_rio(crs):
    from odc.geo import CRS

    if crs.is_epsg_code:
        return CRS(f"EPSG:{crs.to_epsg()}")
    return CRS(crs.wkt)


def rio_geobox(meta):
    """Construct geobox from src.meta of opened rasterio dataset"""
    if "crs" not in meta or "transform" not in meta:
        return None

    h, w = (meta["height"], meta["width"])
    crs = dc_crs_from_rio(meta["crs"])
    transform = meta["transform"]

    return GeoBox(wh_(w, h), transform, crs)


def _fix_resampling(kw: dict) -> None:
    r = kw.get("resampling")
    if isinstance(r, str):
        kw["resampling"] = resampling_s2rio(r)


def rio_slurp_reproject(fname, geobox, dtype=None, dst_nodata=None, **kw):
    """
    Read image with reprojection
    """
    import rasterio
    from rasterio.warp import reproject

    _fix_resampling(kw)

    with rasterio.open(str(fname), "r") as src:
        if src.count == 1:
            shape = geobox.shape
            src_band = rasterio.band(src, 1)
        else:
            shape = (src.count, *geobox.shape)
            src_band = rasterio.band(src, tuple(range(1, src.count + 1)))

        if dtype is None:
            dtype = src.dtypes[0]
        if dst_nodata is None:
            dst_nodata = src.nodata
        if dst_nodata is None:
            dst_nodata = 0

        pix = np.full(shape, dst_nodata, dtype=dtype)

        reproject(
            src_band,
            pix,
            dst_nodata=dst_nodata,
            dst_transform=geobox.transform,
            dst_crs=str(geobox.crs),
            **kw,
        )

        meta = src.meta
        meta["src_geobox"] = rio_geobox(meta)
        meta["path"] = fname
        meta["geobox"] = geobox

        return pix, SimpleNamespace(**meta)


def rio_slurp_read(fname, out_shape=None, **kw):
    """
    Read whole image file using rasterio.

    :returns: ndarray (2d or 3d if multi-band), dict (rasterio meta)
    """
    import rasterio

    _fix_resampling(kw)

    if out_shape is not None:
        kw.update(out_shape=out_shape)

    with rasterio.open(str(fname), "r") as src:
        data = src.read(1, **kw) if src.count == 1 else src.read(**kw)
        meta = src.meta
        src_geobox = rio_geobox(meta)

        same_geobox = out_shape is None or out_shape == src_geobox.shape
        geobox = src_geobox if same_geobox else zoom_to(src_geobox, out_shape)

        meta["src_geobox"] = src_geobox
        meta["geobox"] = geobox
        meta["path"] = fname
        return data, SimpleNamespace(**meta)


def rio_slurp(fname, *args, **kw):
    """
    Dispatches to either:

    rio_slurp_read(fname, out_shape, ..)
    rio_slurp_reproject(fname, geobox, ...)
    """
    if len(args) == 0:
        if "geobox" in kw:
            return rio_slurp_reproject(fname, **kw)
        else:
            return rio_slurp_read(fname, **kw)

    if isinstance(args[0], GeoBox):
        return rio_slurp_reproject(fname, *args, **kw)
    else:
        return rio_slurp_read(fname, *args, **kw)


def rio_slurp_xarray(fname, *args, rgb: str = "auto", **kw) -> xr.DataArray:
    """
    Dispatches to either:

    rio_slurp_read(fname, out_shape, ..)
    rio_slurp_reproject(fname, geobox, ...)

    then wraps it all in xr.DataArray with .crs,.nodata etc.
    """

    if len(args) == 0:
        if "geobox" in kw:
            im, mm = rio_slurp_reproject(fname, **kw)
        else:
            im, mm = rio_slurp_read(fname, **kw)
    else:
        if isinstance(args[0], GeoBox):
            im, mm = rio_slurp_reproject(fname, *args, **kw)
        else:
            im, mm = rio_slurp_read(fname, *args, **kw)

    if im.ndim == 3:
        dims = ("band", *mm.geobox.dims)
        if rgb and im.shape[0] in (3, 4):
            im = im.transpose([1, 2, 0])
            dims = tuple(dims[i] for i in [1, 2, 0])
    else:
        dims = mm.geobox.dims

    return xr.DataArray(
        im, dims=dims, coords=xr_coords(mm.geobox), attrs={"nodata": mm.nodata}
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import csv
import json
import logging
import signal
import sys

import click
import pandas as pd
import yaml
import yaml.resolver
from click import echo, style

from datacube.index import Index
from datacube.ui import click as ui
from datacube.ui.click import cli, exit_on_empty_file, print_help_msg
from datacube.utils import InvalidDocException, read_documents
from datacube.utils.serialise import SafeDatacubeDumper

_LOG: logging.Logger = logging.getLogger("datacube-product")


@cli.group(name="product", help="Product commands")
def product_cli() -> None:
    pass


@product_cli.command("add")
@click.option(
    "--allow-exclusive-lock/--forbid-exclusive-lock",
    is_flag=True,
    default=False,
    help="Allow index to be locked from other users while updating (default: false)",
)
@click.argument("files", type=str, nargs=-1)
@ui.pass_index()
def add_products(index: Index, allow_exclusive_lock: bool, files: list) -> None:
    """
    Add or update products in the generic index.
    """
    if not files:
        print_help_msg(add_products)
        sys.exit(1)

    docs = list(read_documents(*files))
    exit_on_empty_file(docs)

    def on_ctrlc(sig, frame) -> None:
        echo("""Can not abort `product add` without leaving database in bad state.

This operation requires constructing a bunch of indexes and this takes time, the
bigger your database the longer it will take. Just wait a bit.""")

    signal.signal(signal.SIGINT, on_ctrlc)

    for descriptor_path, parsed_doc in docs:
        try:
            type_ = index.products.from_doc(parsed_doc)
            echo(f'Adding "{type_.name}" (this might take a while)', nl=False)
            index.products.add(type_, allow_table_lock=allow_exclusive_lock)
            echo(" DONE")
        except InvalidDocException as e:
            _LOG.exception(e)
            _LOG.error("Invalid product definition: %s", descriptor_path)
            sys.exit(1)


@product_cli.command("update")
@click.option(
    "--allow-unsafe/--forbid-unsafe",
    is_flag=True,
    default=False,
    help="Allow unsafe updates (default: false)",
)
@click.option(
    "--allow-exclusive-lock/--forbid-exclusive-lock",
    is_flag=True,
    default=False,
    help="Allow index to be locked from other users while updating (default: false)",
)
@click.option(
    "--dry-run", "-d", is_flag=True, default=False, help="Check if everything is ok"
)
@click.argument("files", type=str, nargs=-1)
@ui.pass_index()
def update_products(
    index: Index,
    allow_unsafe: bool,
    allow_exclusive_lock: bool,
    dry_run: bool,
    files: list,
) -> None:
    """
    Update existing products.

    An error will be thrown if a change is potentially unsafe.

    (An unsafe change is anything that may potentially make the product
    incompatible with existing datasets of that type)
    """
    if not files:
        print_help_msg(update_products)
        sys.exit(1)

    docs = list(read_documents(*files))
    exit_on_empty_file(docs)
    failures = 0
    for descriptor_path, parsed_doc in docs:
        try:
            type_ = index.products.from_doc(parsed_doc)
        except InvalidDocException as e:
            _LOG.exception(e)
            _LOG.error("Invalid product definition: %s", descriptor_path)
            failures += 1
            continue

        if not dry_run:
            try:
                index.products.update(
                    type_,
                    allow_unsafe_updates=allow_unsafe,
                    allow_table_lock=allow_exclusive_lock,
                )
                echo(f'Updated "{type_.name}"')
            except ValueError as e:
                echo(f'Failed to update "{type_.name}": {e}')
                failures += 1
        else:
            can_update, safe_changes, unsafe_changes = index.products.can_update(
                type_, allow_unsafe_updates=allow_unsafe
            )

            if can_update:
                echo(
                    f'Can update "{type_.name}": {len(list(unsafe_changes))} unsafe '
                    f"changes, {len(list(safe_changes))} safe changes"
                )
            else:
                echo(
                    f'Cannot update "{type_.name}": {len(list(unsafe_changes))} unsafe'
                    f"changes, {len(list(safe_changes))} safe changes"
                )
    sys.exit(failures)


@product_cli.command("delete", help="Delete products and all associated datasets")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Allow a product with active datasets to be deleted (default: false)",
)
@click.option(
    "--dry-run", "-d", is_flag=True, default=False, help="Check if everything is ok"
)
@click.argument("product_names", type=str, nargs=-1)
@ui.pass_index()
def delete_products(
    index: Index, force: bool, dry_run: bool, product_names: list
) -> None:
    """
    Delete products.

    An error will be thrown if the product has active datasets, unless the force option is provided.
    """
    if not product_names:
        print_help_msg(delete_products)
        sys.exit(1)

    try:
        products = [index.products.get_by_name_unsafe(name) for name in product_names]
    except KeyError as e:
        click.echo(str(e), err=True)
        sys.exit(1)

    if sys.stdin.isatty() and force:
        click.confirm(
            "Warning: you may be deleting active datasets. Proceed?", abort=True
        )

    if not dry_run:
        deleted = index.products.delete(products, force)
        not_deleted = set(product_names).difference({p.name for p in deleted})
        if not force and not_deleted:
            click.echo(
                f"Product(s) {', '.join(not_deleted)} could not be deleted due to active datasets. "
                "Use the --force option to delete anyway."
            )
        click.echo(
            f"{len(deleted)} out of {len(product_names)} products successfully deleted."
        )
    else:
        click.echo(f"{len(products)} products not deleted (dry run)")

    click.echo("Completed product deletion.")


def _write_csv(products) -> None:
    product_dicts = [prod.to_dict() for prod in products]
    writer = csv.DictWriter(
        sys.stdout,
        [
            "id",
            "name",
            "description",
            "ancillary_quality",
            "latgqa_cep90",
            "product_type",
            "gqa_abs_iterative_mean_xy",
            "gqa_ref_source",
            "sat_path",
            "gqa_iterative_stddev_xy",
            "time",
            "sat_row",
            "orbit",
            "gqa",
            "instrument",
            "gqa_abs_xy",
            "crs",
            "resolution",
            "tile_shape",
            "spatial_dimensions",
        ],
        extrasaction="ignore",
    )
    writer.writeheader()
    writer.writerows(product_dicts)


def _write_yaml(products: list):
    """
    Dump yaml data with support for OrderedDicts.

    Allows for better human-readability of output: such as dataset ID field first, sources last.

    (Ordered dicts are output identically to normal yaml dicts: their order is purely for readability)
    """
    product_dicts = [prod.to_dict() for prod in products]

    return yaml.dump_all(
        product_dicts,
        sys.stdout,
        Dumper=SafeDatacubeDumper,
        default_flow_style=False,
        indent=4,
    )


def _write_tab(products: list) -> None:
    df = pd.DataFrame(prod.to_dict() for prod in products)

    if df.empty:
        echo("No products discovered :(")
        return

    output_columns = (
        "id",
        "name",
        "description",
        "ancillary_quality",
        "product_type",
        "gqa_abs_iterative_mean_xy",
        "gqa_ref_source",
        "sat_path",
        "gqa_iterative_stddev_xy",
        "time",
        "sat_row",
        "orbit",
        "gqa",
        "instrument",
        "gqa_abs_xy",
        "crs",
        "resolution",
        "tile_shape",
        "spatial_dimensions",
    )
    # If the intersection of desired columns with available columns is empty, just use whatever IS in df
    output_columns = (
        tuple(col for col in output_columns if col in df.columns) or df.columns
    )
    echo(df.to_string(columns=output_columns, justify="left", index=False))


def _default_lister(products) -> None:
    products = list(products)
    if len(products) == 0:
        return

    max_w = max(len(p.name) for p in products)

    for prod in products:
        name = "{s:<{n}}".format(s=prod.name, n=max_w)
        echo(style(name, fg="green") + "  " + prod.definition.get("description", ""))


LIST_OUTPUT_WRITERS = {
    "default": _default_lister,
    "csv": _write_csv,
    "yaml": _write_yaml,
    "tab": _write_tab,
}


@product_cli.command("list")
@click.option(
    "-f",
    "output_format",
    help="Output format",
    type=click.Choice(list(LIST_OUTPUT_WRITERS)),
    default="default",
    show_default=True,
)
@ui.pass_datacube()
def list_products(dc, output_format) -> None:
    """
    List products that are defined in the generic index.
    """
    products = dc.index.products.search()

    writer = LIST_OUTPUT_WRITERS[output_format]

    writer(products)


@product_cli.command("show")
@click.option(
    "-f",
    "output_format",
    help="Output format",
    type=click.Choice(["yaml", "json"]),
    default="yaml",
    show_default=True,
)
@click.argument("product_name", nargs=-1)
@ui.pass_datacube()
def show_product(dc, product_name: str, output_format: str) -> None:
    """
    Show details about a product in the generic index.
    """
    if len(product_name) == 0:
        products = list(dc.index.products.get_all())
    else:
        products = []
        for name in product_name:
            p = dc.index.products.get_by_name(name)
            if p is None:
                echo(f"No such product: {name!r}", err=True)
                sys.exit(1)
            else:
                products.append(p)

    if len(products) == 0:
        echo("No products", err=True)
        sys.exit(1)

    if output_format == "yaml":
        yaml.dump_all(
            (p.definition for p in products),
            sys.stdout,
            Dumper=SafeDatacubeDumper,
            default_flow_style=False,
            indent=4,
        )
    elif output_format == "json":
        if len(products) > 1:
            echo("Can not output more than 1 product in json format", err=True)
            sys.exit(1)
        product, *_ = products
        click.echo_via_pager(json.dumps(product.definition, indent=4))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Iterable

from typing_extensions import override

from datacube.index.abstract import AbstractUserResource


class UserResource(AbstractUserResource):
    def __init__(self) -> None:
        pass

    @override
    def grant_role(self, role: str, *usernames: str) -> None:
        raise NotImplementedError()

    @override
    def create_user(
        self, username: str, password: str, role: str, description: str | None = None
    ) -> None:
        raise NotImplementedError()

    @override
    def delete_user(self, *usernames: str) -> None:
        raise NotImplementedError()

    @override
    def list_users(self) -> Iterable[tuple[str, str, str]]:
        return []



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Test database methods.

Integration tests: these depend on a local Postgres instance.
"""

import copy
import datetime
from pathlib import Path
from uuid import UUID

import pytest

from datacube.index import Index
from datacube.index.eo3 import prep_eo3
from datacube.index.exceptions import MissingRecordError
from datacube.model import Dataset, MetadataType, Product
from datacube.testutils import suppress_deprecations
from datacube.utils.json_types import JsonDict

_telemetry_uuid = UUID("4ec8fe97-e8b9-11e4-87ff-1040f381a756")
_telemetry_dataset = {
    "product_type": "satellite_telemetry_data",
    "checksum_path": "package.sha1",
    "id": str(_telemetry_uuid),
    "ga_label": "LS8_OLITIRS_STD-MD_P00_LC81160740742015089ASA00_"
    "116_074_20150330T022553Z20150330T022657",
    "ga_level": "P00",
    "size_bytes": 637660782,
    "platform": {"code": "LANDSAT_8"},
    # We're unlikely to have extent info for a raw dataset, we'll use it for search tests.
    "extent": {
        "center_dt": datetime.datetime(2014, 7, 26, 23, 49, 0, 343853).isoformat(),
        "coord": {
            "ll": {"lat": -31.33333, "lon": 149.78434},
            "lr": {"lat": -31.37116, "lon": 152.20094},
            "ul": {"lat": -29.23394, "lon": 149.85216},
            "ur": {"lat": -29.26873, "lon": 152.21782},
        },
    },
    "creation_dt": datetime.datetime(2015, 4, 22, 6, 32, 4).isoformat(),
    "instrument": {"name": "OLI_TIRS"},
    "format": {"name": "MD"},
    "lineage": {"source_datasets": {}, "blah": float("NaN")},
}

_pseudo_telemetry_dataset_type: JsonDict = {
    "name": "ls8_telemetry",
    "description": "LS8 test",
    "metadata": {
        "product_type": "satellite_telemetry_data",
        "platform": {"code": "LANDSAT_8"},
        "format": {"name": "MD"},
    },
    "metadata_type": "eo",
}


def test_archive_datasets(index: Index, ls8_eo3_dataset) -> None:
    datasets = list(index.datasets.search())
    assert len(datasets) == 1
    assert not datasets[0].is_archived

    index.datasets.archive([ls8_eo3_dataset.id])
    datasets = list(index.datasets.search())
    assert len(datasets) == 0

    # The model should show it as archived now.
    indexed_dataset = index.datasets.get(ls8_eo3_dataset.id)
    assert indexed_dataset is not None
    assert indexed_dataset.is_archived

    index.datasets.restore([ls8_eo3_dataset.id])
    datasets = list(index.datasets.search())
    assert len(datasets) == 1

    # And now active
    indexed_dataset = index.datasets.get(ls8_eo3_dataset.id)
    assert indexed_dataset is not None
    assert not indexed_dataset.is_archived


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_archive_less_mature(
    index: Index, final_dataset, nrt_dataset, ds_no_region
) -> None:
    # case 1: add nrt then final; nrt should get archived
    index.datasets.add(nrt_dataset, with_lineage=False, archive_less_mature=True)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    index.datasets.add(final_dataset, with_lineage=False, archive_less_mature=True)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert d.is_archived
    d = index.datasets.get(final_dataset.id)
    assert d is not None
    assert not d.is_archived

    # case 2: purge nrt; re-add with final already there
    index.datasets.purge([nrt_dataset.id])
    assert index.datasets.get(nrt_dataset.id) is None
    with pytest.raises(ValueError):
        # should error as more mature version of dataset already exists
        index.datasets.add(nrt_dataset, with_lineage=False, archive_less_mature=True)

    # case 3: re-index final; nrt should still get archived
    assert index.datasets.get(final_dataset.id) is not None
    index.datasets.add(nrt_dataset, with_lineage=False)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    index.datasets.add(final_dataset, with_lineage=False, archive_less_mature=True)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert d.is_archived


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_cannot_search_for_less_mature(index: Index, nrt_dataset, ds_no_region) -> None:
    # if a dataset is missing a property required for finding less mature datasets,
    # it should error
    index.datasets.add(nrt_dataset, with_lineage=False, archive_less_mature=0)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    assert ds_no_region.metadata.region_code is None
    with pytest.raises(ValueError, match="region_code"):
        index.datasets.add(ds_no_region, with_lineage=False, archive_less_mature=0)


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_archive_less_mature_approx_timestamp(
    index: Index, ga_s2am_ard3_final, ga_s2am_ard3_interim
) -> None:
    # test archive_less_mature where there's a slight difference in timestamps
    index.datasets.add(ga_s2am_ard3_interim, with_lineage=False)
    d = index.datasets.get(ga_s2am_ard3_interim.id)
    assert d is not None
    assert not d.is_archived
    index.datasets.add(ga_s2am_ard3_final, with_lineage=False, archive_less_mature=True)
    d = index.datasets.get(ga_s2am_ard3_interim.id)
    assert d is not None
    assert d.is_archived
    d = index.datasets.get(ga_s2am_ard3_final.id)
    assert d is not None
    assert not d.is_archived


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_dont_archive_less_mature(index: Index, final_dataset, nrt_dataset) -> None:
    # ensure datasets aren't archive if no archive_less_mature value is provided
    index.datasets.add(nrt_dataset, with_lineage=False)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    index.datasets.add(final_dataset, with_lineage=False, archive_less_mature=None)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    d = index.datasets.get(final_dataset.id)
    assert d is not None
    assert not d.is_archived


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_archive_less_mature_bool(index: Index, final_dataset, nrt_dataset) -> None:
    # if archive_less_mature value gets passed as a bool via an outdated script
    index.datasets.add(nrt_dataset, with_lineage=False)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    index.datasets.add(final_dataset, with_lineage=False, archive_less_mature=False)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    d = index.datasets.get(final_dataset.id)
    assert d is not None
    assert not d.is_archived


def test_purge_datasets(index: Index, ls8_eo3_dataset) -> None:
    assert index.datasets.has(ls8_eo3_dataset.id)
    datasets = list(index.datasets.search())
    assert len(datasets) == 1
    assert not datasets[0].is_archived

    # Archive dataset
    index.datasets.archive([ls8_eo3_dataset.id])
    datasets = list(index.datasets.search())
    assert len(datasets) == 0

    # The model should show it as archived now.
    indexed_dataset = index.datasets.get(ls8_eo3_dataset.id)
    assert indexed_dataset is not None
    assert indexed_dataset.is_archived

    # Purge dataset
    index.datasets.purge([ls8_eo3_dataset.id])
    assert index.datasets.get(ls8_eo3_dataset.id) is None


def test_purge_datasets_cli(index: Index, ls8_eo3_dataset, clirunner) -> None:
    dsid = ls8_eo3_dataset.id

    # Attempt to purge non-archived dataset should fail
    runner = clirunner(["dataset", "purge", str(dsid)])
    assert "could not be purged" in runner.output
    assert str(dsid) in runner.output
    assert "0 of 1 datasets purged" in runner.output

    # Archive dataset
    index.datasets.archive([dsid])
    indexed_dataset = index.datasets.get(dsid)
    assert indexed_dataset is not None
    assert indexed_dataset.is_archived

    # Test CLI dry run
    clirunner(["dataset", "purge", "--dry-run", str(dsid)])
    indexed_dataset = index.datasets.get(dsid)
    assert indexed_dataset is not None
    assert indexed_dataset.is_archived

    # Test CLI purge
    clirunner(["dataset", "purge", str(dsid)])
    assert index.datasets.get(dsid) is None

    # Attempt to purge non-existent dataset should fail
    clirunner(["dataset", "purge", str(dsid)], expect_success=False)


def test_purge_all_datasets_cli(
    index: Index, cfg_env, ls8_eo3_dataset, clirunner
) -> None:
    dsid = ls8_eo3_dataset.id

    # archive all datasets
    clirunner(["dataset", "archive", "--all"])

    indexed_dataset = index.datasets.get(dsid)
    assert indexed_dataset is not None
    assert indexed_dataset.is_archived

    # Restore all datasets
    clirunner(["dataset", "restore", "--all"])
    indexed_dataset = index.datasets.get(dsid)
    assert indexed_dataset is not None
    assert not indexed_dataset.is_archived

    # Archive again
    clirunner(["dataset", "archive", "--all"])

    # and purge
    clirunner(["dataset", "purge", "--all"])
    assert index.datasets.get(dsid) is None


def test_index_duplicate_dataset(index: Index, cfg_env, ls8_eo3_dataset) -> None:
    product = ls8_eo3_dataset.product
    dsid = ls8_eo3_dataset.id
    assert index.datasets.has(dsid)

    # Insert again.
    ds = Dataset(product, ls8_eo3_dataset.metadata_doc, uri=ls8_eo3_dataset.uri)
    index.datasets.add(ds, with_lineage=False)

    assert index.datasets.has(dsid)


def test_has_dataset(index: Index, ls8_eo3_dataset: Dataset) -> None:
    assert index.datasets.has(ls8_eo3_dataset.id)
    assert index.datasets.has(str(ls8_eo3_dataset.id))

    assert not index.datasets.has(UUID("f226a278-e422-11e6-b501-185e0f80a5c0"))
    assert not index.datasets.has("f226a278-e422-11e6-b501-185e0f80a5c0")

    assert index.datasets.bulk_has(
        [ls8_eo3_dataset.id, UUID("f226a278-e422-11e6-b501-185e0f80a5c0")]
    ) == [True, False]
    assert index.datasets.bulk_has(
        [str(ls8_eo3_dataset.id), "f226a278-e422-11e6-b501-185e0f80a5c0"]
    ) == [True, False]


def test_get_dataset(index: Index, ls8_eo3_dataset: Dataset) -> None:
    assert index.datasets.has(ls8_eo3_dataset.id)
    assert index.datasets.has(str(ls8_eo3_dataset.id))

    assert index.datasets.bulk_has(
        [ls8_eo3_dataset.id, "f226a278-e422-11e6-b501-185e0f80a5c0"]
    ) == [True, False]

    for tr in (lambda x: x, lambda x: str(x)):
        ds = index.datasets.get(tr(ls8_eo3_dataset.id))
        assert ds is not None
        assert ds.id == ls8_eo3_dataset.id

        (ds,) = index.datasets.bulk_get([tr(ls8_eo3_dataset.id)])
        assert ds.id == ls8_eo3_dataset.id

    assert (
        index.datasets.bulk_get(
            [
                "f226a278-e422-11e6-b501-185e0f80a5c0",
                "f226a278-e422-11e6-b501-185e0f80a5c1",
            ]
        )
        == []
    )


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_add_dataset_no_product_id(
    index: Index, extended_eo3_metadata_type, ls8_eo3_product, eo3_ls8_dataset_doc
) -> None:
    product_no_id = Product(extended_eo3_metadata_type, ls8_eo3_product.definition)
    assert product_no_id.id is None
    dataset_doc, _ = eo3_ls8_dataset_doc
    dataset = Dataset(product_no_id, prep_eo3(dataset_doc))
    assert index.datasets.add(dataset, with_lineage=False)


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_transactions_api_ctx_mgr(
    index: Index,
    extended_eo3_metadata_type_doc,
    ls8_eo3_product,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
):
    from datacube.index.hl import Doc2Dataset

    resolver = Doc2Dataset(index, products=[ls8_eo3_product.name], verify_lineage=False)
    ds1, err = resolver(*eo3_ls8_dataset_doc)
    assert ds1 is not None
    ds2, err = resolver(*eo3_ls8_dataset2_doc)
    assert ds2 is not None
    with pytest.raises(Exception) as e:  # noqa: SIM117
        with index.transaction() as trans:
            assert index.datasets.get(ds1.id) is None
            index.datasets.add(ds1, with_lineage=False)
            assert index.datasets.get(ds1.id) is not None
            raise Exception("Rollback!")
    assert "Rollback!" in str(e.value)
    assert index.datasets.get(ds1.id) is None
    with index.transaction() as trans:
        assert index.datasets.get(ds1.id) is None
        index.datasets.add(ds1, with_lineage=False)
        assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds1.id) is not None
    with index.transaction() as trans:
        index.datasets.add(ds2, with_lineage=False)
        assert index.datasets.get(ds2.id) is not None
        raise trans.rollback_exception("Rollback")
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is None


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_transactions_api_ctx_mgr_nested(
    index: Index,
    extended_eo3_metadata_type_doc,
    ls8_eo3_product,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
):
    from datacube.index.hl import Doc2Dataset

    resolver = Doc2Dataset(index, products=[ls8_eo3_product.name], verify_lineage=False)
    ds1, err = resolver(*eo3_ls8_dataset_doc)
    assert ds1 is not None
    ds2, err = resolver(*eo3_ls8_dataset2_doc)
    assert ds2 is not None
    with pytest.raises(Exception) as e:  # noqa: SIM117
        with index.transaction():
            with index.transaction() as trans:
                assert index.datasets.get(ds1.id) is None
                index.datasets.add(ds1, False)
                assert index.datasets.get(ds1.id) is not None
                raise Exception("Rollback!")
    assert "Rollback!" in str(e.value)
    assert index.datasets.get(ds1.id) is None
    with index.transaction():  # noqa: SIM117
        with index.transaction() as trans:
            assert index.datasets.get(ds1.id) is None
            index.datasets.add(ds1, False)
            assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds1.id) is not None
    with index.transaction():  # noqa: SIM117
        with index.transaction() as trans:
            index.datasets.add(ds2, False)
            assert index.datasets.get(ds2.id) is not None
            raise trans.rollback_exception("Rollback")
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is None


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_transactions_api_manual(
    index: Index,
    extended_eo3_metadata_type_doc,
    ls8_eo3_product,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
) -> None:
    from datacube.index.hl import Doc2Dataset

    resolver = Doc2Dataset(index, products=[ls8_eo3_product.name], verify_lineage=False)
    ds1, err = resolver(*eo3_ls8_dataset_doc)
    ds2, err = resolver(*eo3_ls8_dataset2_doc)
    assert ds1 is not None
    assert ds2 is not None
    trans = index.transaction()
    index.datasets.add(ds1, False)
    assert index.datasets.get(ds1.id) is not None
    trans.begin()
    index.datasets.add(ds2, False)
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is not None
    trans.rollback()
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is None
    trans.begin()
    index.datasets.add(ds2, False)
    trans.commit()
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is not None


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_transactions_api_hybrid(
    index: Index,
    extended_eo3_metadata_type_doc,
    ls8_eo3_product,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
) -> None:
    from datacube.index.hl import Doc2Dataset

    resolver = Doc2Dataset(index, products=[ls8_eo3_product.name], verify_lineage=False)
    ds1, err = resolver(*eo3_ls8_dataset_doc)
    assert ds1 is not None
    ds2, err = resolver(*eo3_ls8_dataset2_doc)
    assert ds2 is not None
    with index.transaction() as trans:
        assert index.datasets.get(ds1.id) is None
        index.datasets.add(ds1, False)
        assert index.datasets.get(ds1.id) is not None
        trans.rollback()
        assert index.datasets.get(ds1.id) is None
        trans.begin()
        assert index.datasets.get(ds1.id) is None
        index.datasets.add(ds1, False)
        assert index.datasets.get(ds1.id) is not None
        trans.commit()
        assert index.datasets.get(ds1.id) is not None
        trans.begin()
        index.datasets.add(ds2, False)
        assert index.datasets.get(ds2.id) is not None
        trans.rollback()
    assert index.datasets.get(ds1.id) is not None
    assert index.datasets.get(ds2.id) is None


def test_get_missing_things(index: Index) -> None:
    """
    The get(id) methods should return None if the object doesn't exist.
    """
    uuid_ = UUID("18474b58-c8a6-11e6-a4b3-185e0f80a5c0")
    missing_thing = index.datasets.get(uuid_, include_sources=False)
    assert missing_thing is None, "get() should return none when it doesn't exist"

    if index.supports_lineage and not index.supports_external_lineage:
        missing_thing = index.datasets.get(uuid_, include_sources=True)
        assert missing_thing is None, "get() should return none when it doesn't exist"

    # Max SmallInteger: https://www.postgresql.org/docs/17/datatype-numeric.html
    id_ = 32767
    missing_thing = index.metadata_types.get(id_)
    assert missing_thing is None, "get() should return none when it doesn't exist"

    missing_thing = index.products.get(id_)
    assert missing_thing is None, "get() should return none when it doesn't exist"


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_index_dataset_with_sources(index: Index, default_metadata_type) -> None:
    type_ = index.products.add_document(_pseudo_telemetry_dataset_type)

    parent_doc = _telemetry_dataset.copy()
    parent = Dataset(type_, parent_doc, None, sources={})
    child_doc = _telemetry_dataset.copy()
    child_doc["lineage"] = {"source_datasets": {"source": _telemetry_dataset}}
    child_doc["id"] = "051a003f-5bba-43c7-b5f1-7f1da3ae9cfb"
    child = Dataset(type_, child_doc, sources={"source": parent})

    with pytest.raises(MissingRecordError):
        index.datasets.add(child, with_lineage=False)

    index.datasets.add(child)
    assert index.datasets.get(parent.id)
    assert index.datasets.get(child.id)

    assert len(list(index.datasets.bulk_get([parent.id, child.id]))) == 2

    index.datasets.add(child, with_lineage=False)
    index.datasets.add(child, with_lineage=True)

    parent_doc["platform"] = {"code": "LANDSAT_9"}
    index.datasets.add(child, with_lineage=True)
    index.datasets.add(child, with_lineage=False)


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_index_dataset_with_lineage(
    index: Index, ds_with_lineage, ls8_eo3_dataset
) -> None:
    assert ds_with_lineage.source_tree
    index.datasets.add(ds_with_lineage)
    sources = index.lineage.get_source_tree(ds_with_lineage.id).children
    assert sources is not None
    assert len(sources["ard"]) == 1
    assert sources["ard"][0].dataset_id == ls8_eo3_dataset.id
    assert index.datasets.get(ds_with_lineage.id)


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_index_dataset_with_location(
    index: Index, default_metadata_type: MetadataType
) -> None:
    first_file = Path("/tmp/first/something.yaml").absolute()
    second_file = Path("/tmp/second/something.yaml").absolute()

    product = index.products.add_document(_pseudo_telemetry_dataset_type)
    assert product is not None
    dataset = Dataset(product, _telemetry_dataset, uri=first_file.as_uri(), sources={})
    index.datasets.add(dataset)
    stored = index.datasets.get(dataset.id)
    assert stored is not None
    assert stored.id == _telemetry_uuid
    # TODO: Dataset types?
    assert stored.product.id == product.id
    assert stored.metadata_type.id == default_metadata_type.id
    assert stored.local_path == Path(first_file)

    # Ingesting again should have no effect.
    index.datasets.add(dataset)
    stored = index.datasets.get(dataset.id)
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 1
    # Remove the location
    with suppress_deprecations():
        was_removed = index.datasets.remove_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_removed
    with suppress_deprecations():
        was_removed = index.datasets.remove_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert not was_removed
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 0
    # Re-add the location
    with suppress_deprecations():
        was_added = index.datasets.add_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_added
    with suppress_deprecations():
        was_added = index.datasets.add_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert not was_added
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 1

    # A rough date is ok: 1:01 beforehand just in case someone runs this during daylight savings time conversion :)
    # (any UTC conversion errors will be off by much more than this for PST/AEST)
    before_archival_dt = utc_now() - datetime.timedelta(hours=1, minutes=1)

    with suppress_deprecations():
        was_archived = index.datasets.archive_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_archived
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == []
    with suppress_deprecations():
        locations = index.datasets.get_archived_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == [first_file.as_uri()]

    # It should return the time archived.
    with suppress_deprecations():
        location_times = index.datasets.get_archived_location_times(
            dataset.id
        )  # Test of deprecated method
    assert len(location_times) == 1
    location, archived_time = location_times[0]
    assert location == first_file.as_uri()
    assert utc_now() > archived_time > before_archival_dt

    with suppress_deprecations():
        was_restored = index.datasets.restore_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_restored
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 1

    # Indexing with a new path should NOT add the second one.
    dataset._uris = [second_file.as_uri()]
    index.datasets.add(dataset)
    stored = index.datasets.get(dataset.id)
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 1

    # Add location manually instead
    with suppress_deprecations():
        index.datasets.add_location(
            dataset.id, second_file.as_uri()
        )  # Test of deprecated method
        stored = index.datasets.get(dataset.id)
        assert stored is not None
    assert len(stored._uris) == 2

    # Newest to oldest.
    assert stored._uris == [second_file.as_uri(), first_file.as_uri()]
    # And the second one is newer, so it should be returned as the default local path:
    assert stored.local_path == Path(second_file)

    # Can archive and restore the first file, and location order is preserved
    with suppress_deprecations():
        was_archived = index.datasets.archive_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_archived
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == [second_file.as_uri()]
    with suppress_deprecations():
        was_restored = index.datasets.restore_location(
            dataset.id, first_file.as_uri()
        )  # Test of deprecated method
    assert was_restored
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == [second_file.as_uri(), first_file.as_uri()]

    # Can archive and restore the second file, and location order is preserved
    with suppress_deprecations():
        was_archived = index.datasets.archive_location(
            dataset.id, second_file.as_uri()
        )  # Test of deprecated method
    assert was_archived
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == [first_file.as_uri()]
    with suppress_deprecations():
        was_restored = index.datasets.restore_location(
            dataset.id, second_file.as_uri()
        )  # Test of deprecated method
    assert was_restored
    with suppress_deprecations():
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert locations == [second_file.as_uri(), first_file.as_uri()]

    # Indexing again without location should have no effect.
    dataset._uris = []
    with suppress_deprecations():
        index.datasets.add(dataset)
        stored = index.datasets.get(dataset.id)
        assert stored is not None
        locations = index.datasets.get_locations(
            dataset.id
        )  # Test of deprecated method
    assert len(locations) == 2
    # Newest to oldest.
    assert locations == [second_file.as_uri(), first_file.as_uri()]
    # And the second one is newer, so it should be returned as the default local path:
    assert stored.local_path == Path(second_file)

    # Check order of uris is preserved when indexing with more than one
    second_ds_doc = copy.deepcopy(_telemetry_dataset)
    test_uuid = "366f32d8-e1f8-11e6-94b4-185e0f80a589"
    second_ds_doc["id"] = test_uuid
    with suppress_deprecations():
        index.datasets.add(
            Dataset(  # Test deprecated behaviour
                product, second_ds_doc, uris=["file:///a", "file:///b"], sources={}
            )
        )

    # test order using get_locations function
    with suppress_deprecations():
        # Test of deprecated method
        assert index.datasets.get_locations(test_uuid) == [
            "file:///a",
            "file:///b",
        ]

        # test order using datasets.get(), it has custom query as it turns out
        d = index.datasets.get(test_uuid)
        assert d is not None
        assert d._uris == [
            "file:///a",
            "file:///b",
        ]

        # test update, this should prepend file:///c, file:///d to the existing list
        index.datasets.update(
            Dataset(  # Test of deprecated functionality
                product,
                second_ds_doc,
                uris=["file:///a", "file:///c", "file:///d"],
                sources={},
            )
        )
        assert index.datasets.get_locations(test_uuid) == [  # Test of deprecated method
            "file:///c",
            "file:///d",
            "file:///a",
            "file:///b",
        ]
        d = index.datasets.get(test_uuid)
        assert d is not None
        assert d.uris == [  # Test of deprecated functionality
            "file:///c",
            "file:///d",
            "file:///a",
            "file:///b",
        ]

    # Ability to get datasets for a location
    # Add a second dataset with a different location (to catch lack of joins, filtering etc)
    second_ds_doc = copy.deepcopy(_telemetry_dataset)
    second_ds_doc["id"] = "366f32d8-e1f8-11e6-94b4-185e0f80a5c0"
    index.datasets.add(
        Dataset(product, second_ds_doc, uri=second_file.as_uri(), sources={})
    )
    for mode in ("exact", "prefix", None):
        with suppress_deprecations():
            dataset_ids = [
                d.id
                for d in index.datasets.get_datasets_for_location(
                    first_file.as_uri(), mode=mode
                )
            ]
        assert dataset_ids == [dataset.id]

    assert (
        list(
            index.datasets.get_datasets_for_location(first_file.as_uri() + "#part=100")
        )
        == []
    )

    with pytest.raises(ValueError):
        list(
            index.datasets.get_datasets_for_location(
                first_file.as_uri(), mode="nosuchmode"
            )
        )


def utc_now():
    return datetime.datetime.now(datetime.UTC)


def test_bulk_reads_transaction(
    index,
    extended_eo3_metadata_type_doc,
    ls8_eo3_product,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
) -> None:
    with pytest.raises(ValueError) as e:  # noqa: SIM117
        with index.datasets._db_connection() as conn:
            conn.bulk_simple_dataset_search(batch_size=2)
    assert "within a transaction" in str(e.value)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import uuid
from unittest.mock import MagicMock, patch

import numpy
import pytest
from odc.geo import CRS, Resolution
from odc.geo.gridspec import GridSpec
from typing_extensions import override

from datacube.api.grid_workflow import GridWorkflow, GridWorkflowException
from datacube.testutils import mk_sample_product


class PickleableMock(MagicMock):
    @override
    def __reduce__(self):
        return (MagicMock, ())


def mk_fake_index(products, datasets):
    fakeindex = PickleableMock()
    fakeindex._db = None

    fakeindex.products.get_by_name = lambda name: products.get(name, None)

    fakeindex.datasets.get_field_names.return_value = ["time"]  # permit query on time
    fakeindex.datasets.search_eager.return_value = datasets

    return fakeindex


@pytest.fixture
def fake_index():
    return mk_fake_index(
        products={
            "with_gs": mk_sample_product("with_gs", with_grid_spec=True),
            "without_gs": mk_sample_product("without_gs", with_grid_spec=False),
        },
        datasets=[],
    )


def test_create_gridworkflow_creation_with_product(fake_index):
    index = fake_index

    # need product or grispec
    with pytest.raises(GridWorkflowException):
        GridWorkflow(index)

    # Can't specify a product that doesn't exist.
    with pytest.raises(GridWorkflowException):
        GridWorkflow(index, product="no-such-product")

    # Error out if a production without a GridSpec is specified
    assert fake_index.products.get_by_name("without_gs") is not None
    assert fake_index.products.get_by_name("without_gs").grid_spec is None

    with pytest.raises(GridWorkflowException):
        GridWorkflow(index, product="without_gs")

    # Able to create a Grid Workflow with a Product that has a GridSpec
    product = fake_index.products.get_by_name("with_gs")
    assert product is not None
    assert product.grid_spec is not None
    gw = GridWorkflow(index, product="with_gs")
    assert gw.grid_spec is product.grid_spec


@pytest.fixture
def sample_grid_workflow():
    crs = CRS("EPSG:3577")
    pixel = 10  # square pixel linear dimension in crs units
    grid = 10  # size of a single tile in pixels
    gridspec = GridSpec(crs=crs, tile_shape=(grid, grid), resolution=Resolution(pixel))
    fakedataset = MagicMock()
    fakedataset.extent = gridspec.tile_geobox((1, -2)).extent
    fakedataset.center_time = datetime.datetime(2001, 2, 15)
    fakedataset.id = uuid.uuid4()
    fakeindex = PickleableMock()
    fakeindex._db = None
    fakeindex.datasets.get_field_names.return_value = ["time"]
    fakeindex.products.get_field_names.return_value = ["time"]
    fakeindex.datasets.search_eager.return_value = [fakedataset]
    gw = GridWorkflow(fakeindex, gridspec)
    gw.index = fakeindex  # Need to force the fake index
    return gw, gridspec, fakedataset, fakeindex


def test_gridworkflow_str_repr(sample_grid_workflow):
    gw, _, _, _ = sample_grid_workflow
    assert len(str(gw)) > 0
    assert len(repr(gw)) > 0


def test_gridworkflow_cell_observations(sample_grid_workflow):
    gw, gridspec, _, _ = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    assert list(gw.cell_observations(**query).keys()) == [(1, -2)]
    assert list(
        gw.cell_observations(
            **query, geopolygon=gridspec.tile_geobox((1, -2)).extent
        ).keys()
    ) == [(1, -2)]


def test_gridworkflow_cell_observations_errors(sample_grid_workflow):
    gw, gridspec, _, _ = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    # It's invalid to supply tile_buffer and geopolygon at the same time
    with pytest.raises(GridWorkflowException) as e:
        list(
            gw.cell_observations(
                **query,
                tile_buffer=(1, 1),
                geopolygon=gridspec.tile_geobox((1, -2)).extent,
            ).keys()
        )
    assert str(e.value) == "Cannot process tile_buffering and geopolygon together."


def test_gridworkflow_list_tiles_unpadded(sample_grid_workflow):
    gw, _, _, _ = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    assert len(gw.list_tiles(**query)) == 1


def test_gridworkflow_list_tiles_padded(sample_grid_workflow):
    gw, _, _, _ = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    assert len(gw.list_tiles(tile_buffer=(20, 20), **query)) == 9


def test_gridworkflow_list_tiles_multiple_datasets(sample_grid_workflow):
    gw, gridspec, fakedataset, fakeindex = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }

    # Add dataset to cell (2,-2)
    fakedataset2 = MagicMock()
    fakedataset2.extent = gridspec.tile_geobox((2, -2)).extent
    fakedataset2.center_time = fakedataset.center_time
    fakedataset2.id = uuid.uuid4()
    fakeindex.datasets.search_eager.return_value = [fakedataset, fakedataset2]

    # unpadded
    assert len(gw.list_tiles(**query)) == 2
    np_time = numpy.datetime64(fakedataset.center_time, "ns")
    assert set(gw.list_tiles(**query).keys()) == {(1, -2, np_time), (2, -2, np_time)}

    # padded
    assert len(gw.list_tiles(tile_buffer=(20, 20), **query)) == 12


def test_gridworkflow_returned_tile_properties(sample_grid_workflow):
    gw, gridspec, fakedataset, fakeindex = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    np_time = numpy.datetime64(fakedataset.center_time, "ns")

    # Add dataset to cell (2,-2)
    fakedataset2 = MagicMock()
    fakedataset2.extent = gridspec.tile_geobox((2, -2)).extent
    fakedataset2.center_time = fakedataset.center_time
    fakedataset2.id = uuid.uuid4()
    fakeindex.datasets.search_eager.return_value = [fakedataset, fakedataset2]

    tile = gw.list_tiles(**query)[1, -2, np_time]
    assert tile.shape == (1, 10, 10)
    assert len(str(tile)) > 0
    assert len(repr(tile)) > 0

    padded_tile = gw.list_tiles(tile_buffer=(20, 20), **query)[1, -2, np_time]
    assert padded_tile.shape == (1, 14, 14)

    assert len(tile.sources.isel(time=0).item()) == 1
    assert len(padded_tile.sources.isel(time=0).item()) == 2

    assert tile.geobox.alignment == padded_tile.geobox.alignment
    assert tile.geobox.affine * (0, 0) == padded_tile.geobox.affine * (2, 2)
    assert tile.geobox.affine * (10, 10) == padded_tile.geobox.affine * (10 + 2, 10 + 2)


def test_gridworkflow_loading(sample_grid_workflow):
    gw, _, fakedataset, _ = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    np_time = numpy.datetime64(fakedataset.center_time, "ns")
    tile = gw.list_tiles(**query)[1, -2, np_time]
    padded_tile = gw.list_tiles(tile_buffer=(20, 20), **query)[1, -2, np_time]

    measurement = {"nodata": 0, "dtype": numpy.int32}
    fakedataset.product.lookup_measurements.return_value = {"dummy": measurement}
    fakedataset2 = MagicMock()
    fakedataset2.product = fakedataset.product

    with patch("datacube.api.core.Datacube.load_data") as loader:
        _ = GridWorkflow.load(tile)
        _ = GridWorkflow.load(padded_tile)

    assert loader.call_count == 2

    for (args, kwargs), loadable in zip(loader.call_args_list, [tile, padded_tile]):
        args = list(args)
        assert args[0] is loadable.sources
        assert args[1] is loadable.geobox
        assert next(iter(args[2].values())) is measurement
        assert "resampling" in kwargs


def test_gridworkflow_cell_index_extract(sample_grid_workflow):
    gw, gridspec, fakedataset, fakeindex = sample_grid_workflow
    query = {
        "product": "fake_product_name",
        "time": ("2001-1-1 00:00:00", "2001-3-31 23:59:59"),
    }
    np_time = numpy.datetime64(fakedataset.center_time, "ns")

    # Add dataset to cell (2,-2)
    fakedataset2 = MagicMock()
    fakedataset2.extent = gridspec.tile_geobox((2, -2)).extent
    fakedataset2.center_time = fakedataset.center_time
    fakedataset2.id = uuid.uuid4()
    fakeindex.datasets.search_eager.return_value = [fakedataset, fakedataset2]

    tile = gw.list_tiles(cell_index=(1, -2), **query)
    assert len(tile) == 1
    assert tile[1, -2, np_time].shape == (1, 10, 10)
    assert len(tile[1, -2, np_time].sources.values[0]) == 1

    padded_tile = gw.list_tiles(cell_index=(1, -2), tile_buffer=(20, 20), **query)
    assert len(padded_tile) == 1
    assert padded_tile[1, -2, np_time].shape == (1, 14, 14)
    assert len(padded_tile[1, -2, np_time].sources.values[0]) == 2

    # query without product is not allowed
    with pytest.raises(RuntimeError):
        gw.list_cells(cell_index=(1, -2), time=query["time"])


def test_gridworkflow_with_time_depth():
    """Test GridWorkflow with time series.
    Also test `Tile` methods `split` and `split_by_time`
    """
    crs = CRS("EPSG:4326")

    pixel = 10  # square pixel linear dimension in crs units
    gridspec = GridSpec(crs=crs, tile_shape=(10, 10), resolution=Resolution(pixel))

    def make_fake_datasets(num_datasets: int):
        start_time = datetime.datetime(2001, 2, 15)
        delta = datetime.timedelta(days=16)
        for i in range(num_datasets):
            fakedataset = MagicMock()
            fakedataset.extent = gridspec.tile_geobox((1, -2)).extent
            fakedataset.center_time = start_time + (delta * i)
            yield fakedataset

    fakeindex = PickleableMock()
    fakeindex.datasets.get_field_names.return_value = ["time"]  # permit query on time
    fakeindex.datasets.search_eager.return_value = list(make_fake_datasets(100))

    # ------ test with time dimension ----

    gw = GridWorkflow(fakeindex, gridspec)
    query = {"product": "fake_product_name"}

    cells = gw.list_cells(**query)
    for _, cell in cells.items():
        #  test Tile.split()
        for _, tile in cell.split("time"):
            assert tile.shape == (1, 10, 10)

        #  test Tile.split_by_time()
        for year, year_cell in cell.split_by_time(freq="A"):
            for t in year_cell.sources.time.values:
                assert str(t)[:4] == year



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""threads related stuff"""

from concurrent.futures import Future
from functools import partial


class FakeThreadPoolExecutor:
    """Limited version of ThreadPool that executes in the current thread."""

    def submit(self, fn, *args, **kwargs) -> Future:
        f: Future = Future()
        try:
            f.set_result(fn(*args, **kwargs))
        except Exception as e:  # pylint: disable=broad-except
            f.set_exception(e)

        return f

    def map(self, fn, *iterables, timeout=None, chunksize: int = 1) -> map:
        return map(partial(self.submit, fn), *iterables)

    def shutdown(self, wait: bool = True) -> None:
        pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from urllib.parse import ParseResult, urlparse

from deprecat import deprecat
from odc.geo import CRS

from datacube.cfg import ODCEnvironment, ODCOptionHandler
from datacube.migration import ODC2DeprecationWarning
from datacube.model import Field, MetadataType
from datacube.utils import cached_property, report_to_user
from datacube.utils.generic import thread_local_cache

from ._datasets import AbstractDatasetResource
from ._lineage import AbstractLineageResource
from ._metadata_types import AbstractMetadataTypeResource
from ._products import AbstractProductResource
from ._transactions import AbstractTransaction
from ._types import DSID, BatchStatus
from ._users import AbstractUserResource

_LOG: logging.Logger = logging.getLogger(__name__)


class AbstractIndex(ABC):
    """
    Abstract base class for an Index.  All Index implementations should
    inherit from this base class, and implement all abstract methods (and
    override other methods and contract flags as required).
    """

    # Interface contracts - implementations should set to True where appropriate.

    # Metadata type support flags
    #   supports legacy ODCv1 EO style metadata types.
    supports_legacy = False
    #   supports eo3 compatible metadata types.
    supports_eo3 = False
    #   supports non-geospatial (e.g. telemetry) metadata types
    supports_nongeo = False
    #   supports geospatial vector (i.e. non-raster) metadata types (reserved for future use)
    supports_vector = False

    # Database/storage feature support flags
    #   supports add() update() remove() etc methods.
    supports_write = False
    #   supports persistent storage. Writes from previous instantiations will persist into future ones.
    #   (Requires supports_write)
    supports_persistance = False
    #    Supports ACID transactions (Requires supports_write)
    supports_transactions = False
    #    Supports per-CRS spatial indexes (Requires supports_write)
    supports_spatial_indexes = False

    # User management support flags
    #   support the index.users API
    supports_users = False

    # Lineage support flags
    #   supports lineage (either legacy or new API)
    supports_lineage = False
    #   supports external lineage API (as described in EP-08).  Requires supports_lineage
    #   IF support_lineage is True and supports_external_lineage is False THEN legacy lineage API.
    supports_external_lineage = False
    #   supports an external lineage home field.  Requires supports_external_lineage
    supports_external_home = False

    @property
    @abstractmethod
    def name(self) -> str:
        """The index name"""

    @property
    @abstractmethod
    def environment(self) -> ODCEnvironment:
        """The cfg.ODCEnvironment object this Index was initialised from."""

    @property
    @abstractmethod
    def url(self) -> str:
        """A string representing the index"""

    @cached_property
    def url_parts(self) -> ParseResult:
        return urlparse(self.url)

    @property
    @abstractmethod
    def users(self) -> AbstractUserResource:
        """A User Resource instance for the index"""

    @property
    @abstractmethod
    def metadata_types(self) -> AbstractMetadataTypeResource:
        """A MetadataType Resource instance for the index"""

    @property
    @abstractmethod
    def products(self) -> AbstractProductResource:
        """A Product Resource instance for the index"""

    @property
    @abstractmethod
    def lineage(self) -> AbstractLineageResource:
        """A Lineage Resource instance for the index"""

    @property
    @abstractmethod
    def datasets(self) -> AbstractDatasetResource:
        """A Dataset Resource instance for the index"""

    @classmethod
    @abstractmethod
    def from_config(
        cls,
        cfg_env: ODCEnvironment,
        application_name: str | None = None,
        validate_connection: bool = True,
    ) -> "AbstractIndex":
        """Instantiate a new index from an ODCEnvironment configuration object"""

    @classmethod
    @abstractmethod
    def get_dataset_fields(cls, doc: dict) -> Mapping[str, Field]:
        """Return dataset search fields from a metadata type document"""

    @abstractmethod
    def init_db(
        self, with_default_types: bool = True, with_permissions: bool = True
    ) -> bool:
        """
        Initialise an empty database.

        :param with_default_types: Whether to create default metadata types
        :param with_permissions: Whether to create db permissions
        :return: true if the database was created, false if already exists
        """

    # Spatial Index API

    def create_spatial_index(self, crs: CRS) -> bool:
        """
        Create a spatial index for a CRS.

        Note that a newly created spatial index is empty.  If there are already datasets in the index whose
        extents can be safely projected into the CRS, then it is necessary to also call update_spatial_index
        otherwise they will not be found by queries against that CRS.

        Only implemented by index drivers with supports_spatial_indexes set to True.

        :param crs: The coordinate reference system to create a spatial index for.
        :return: True if the spatial index was successfully created (or already exists)
        """
        if not self.supports_spatial_indexes:
            raise NotImplementedError(
                "This index driver does not support the Spatial Index API"
            )
        else:
            raise NotImplementedError()

    def spatial_indexes(self, refresh: bool = False) -> Iterable[CRS]:
        """
        Return the CRSs for which spatial indexes have been created.

        :param refresh: If true, query the backend for the list of current spatial indexes.  If false (the default)
                        a cached list of spatial index CRSs may be returned.
        :return: An iterable of CRSs for which spatial indexes exist in the index
        """
        if not self.supports_spatial_indexes:
            raise NotImplementedError(
                "This index driver does not support the Spatial Index API"
            )
        else:
            raise NotImplementedError()

    def update_spatial_index(
        self,
        crses: Sequence[CRS] = [],
        product_names: Sequence[str] = [],
        dataset_ids: Sequence[DSID] = [],
    ) -> int:
        """
        Populate a newly created spatial index (or indexes).

        Spatial indexes are automatically populated with new datasets as they are indexed, but if there were
        datasets already in the index when a new spatial index is created, or if geometries have been added or
        modified outside of the ODC in a populated index (e.g. with SQL) then the spatial indices must be
        updated manually with this method.

        This is a very slow operation.  The product_names and dataset_ids lists can be used to break the
        operation up into chunks or allow faster updating when the spatial index is only relevant to a
        small portion of the entire index.

        :param crses: A list of CRSes whose spatial indexes are to be updated.
                      Default is to update all spatial indexes
        :param product_names: A list of product names to update the spatial indexes.
                              Default is to update for all products
        :param dataset_ids: A list of ids of specific datasets to update in the spatial index.
                            Default is to update for all datasets (or all datasets in the products
                            in the product_names list)
        :return: The number of dataset extents processed - i.e. the number of datasets updated multiplied by the
                 number of spatial indexes updated.
        """
        if not self.supports_spatial_indexes:
            raise NotImplementedError(
                "This index driver does not support the Spatial Index API"
            )
        else:
            raise NotImplementedError()

    def drop_spatial_index(self, crs: CRS) -> bool:
        """
        Remove a spatial index from the database.

        Note that creating spatial indexes on an existing index is a slow and expensive operation.  Do not
        delete spatial indexes unless you are absolutely certain it is no longer required by any users of
        this ODC index.

        :param crs: The CRS whose spatial index is to be deleted.
        :return: True if the spatial index was successfully dropped.
                 False if spatial index could not be dropped.
        """
        if not self.supports_spatial_indexes:
            raise NotImplementedError(
                "This index driver does not support the Spatial Index API"
            )
        else:
            raise NotImplementedError()

    def clone(
        self,
        origin_index: "AbstractIndex",
        batch_size: int = 1000,
        skip_lineage: bool = False,
        lineage_only: bool = False,
    ) -> Mapping[str, BatchStatus]:
        """
        Clone an existing index into this one.

        Steps are:

        1) Clone all metadata types compatible with this index driver.

           - Products and Datasets with incompatible metadata types are excluded from subsequent steps.
           - Existing metadata types are skipped, but products and datasets associated with them are only
             excluded if the existing metadata type does not match the one from the origin index.

        2) Clone all products with "safe" metadata types.

           - Products are included or excluded by metadata type as discussed above.
           - Existing products are skipped, but datasets associated with them are only
             excluded if the existing product definition does not match the one from the origin index.

        3) Clone all datasets with "safe" products

           - Datasets are included or excluded by product and metadata type, as discussed above.
           - Archived datasets and locations are not cloned.

        4) Clone all lineage relations that can be cloned.

           - All lineage relations are skipped if either index driver does not support lineage,
             or if skip_lineage is True.
           - If this index does not support external lineage then lineage relations that reference datasets
             that do not exist in this index after step 3 above are skipped.

        API Note: This API method is not finalised and may be subject to change.

        :param origin_index: Index whose contents we wish to clone.
        :param batch_size: Maximum number of objects to write to the database in one go.
        :param skip_lineage: Skip lineage in cloned result.
        :param lineage_only: Only clone lineage.
        :return: Dictionary containing a BatchStatus named tuple for "metadata_types", "products"
                 and "datasets", and optionally "lineage".
        """
        results = {}
        if not lineage_only:
            if self.supports_spatial_indexes and origin_index.supports_spatial_indexes:
                for crs in origin_index.spatial_indexes(refresh=True):
                    report_to_user(f"Creating spatial index for CRS {crs}")
                    self.create_spatial_index(crs)
                self.update_spatial_index(
                    list(origin_index.spatial_indexes(refresh=False))
                )
            # Clone Metadata Types
            report_to_user("Cloning Metadata Types:")
            results["metadata_types"] = self.metadata_types.bulk_add(
                origin_index.metadata_types.get_all_docs(), batch_size=batch_size
            )
            res = results["metadata_types"]
            msg = (
                f"{res.completed} metadata types loaded ({res.skipped} skipped) in "
                f"{res.seconds_elapsed:.2f}seconds "
                f"({res.completed * 60 / res.seconds_elapsed:.2f} metadata_types/min)"
            )
            report_to_user(msg, logger=_LOG)
            if res.safe:
                metadata_cache = {
                    name: self.metadata_types.get_by_name_unsafe(name)
                    for name in res.safe
                }
            else:
                metadata_cache = {}
            # Clone Products
            report_to_user("Cloning Products:")
            results["products"] = self.products.bulk_add(
                origin_index.products.get_all_docs(),
                metadata_types=metadata_cache,
                batch_size=batch_size,
            )
            res = results["products"]
            msg = (
                f"{res.completed} products loaded ({res.skipped} skipped) in {res.seconds_elapsed:.2f}seconds "
                f"({res.completed * 60 / res.seconds_elapsed:.2f} products/min)"
            )
            report_to_user(msg, logger=_LOG)
            # Clone Datasets (group by product for now for convenience)
            report_to_user("Cloning Datasets:")
            products = (
                [p for p in self.products.get_all() if p.name in res.safe]
                if res.safe
                else []
            )
            results["datasets"] = self.datasets.bulk_add(
                origin_index.datasets.get_all_docs(
                    products=products, batch_size=batch_size
                ),
                batch_size=batch_size,
            )
            res = results["datasets"]
            report_to_user("")
            msg = (
                f"{res.completed} datasets loaded ({res.skipped} skipped) in {res.seconds_elapsed:.2f}seconds "
                f"({res.completed * 60 / res.seconds_elapsed:.2f} datasets/min)"
            )
            report_to_user(msg, logger=_LOG)
        if (
            not self.supports_lineage
            or not origin_index.supports_lineage
            or skip_lineage
        ):
            report_to_user("Skipping lineage")
            return results
        report_to_user("Cloning Lineage:")
        results["lineage"] = self.lineage.bulk_add(
            origin_index.lineage.get_all_lineage(batch_size), batch_size
        )
        res = results["lineage"]
        report_to_user("")
        msg = (
            f"{res.completed} lineage relations loaded ({res.skipped} skipped) in {res.seconds_elapsed:.2f}seconds "
            f"({res.completed * 60 / res.seconds_elapsed:.2f} lineage relations/min)"
        )
        report_to_user(msg, logger=_LOG)
        return results

    @abstractmethod
    def close(self) -> None:
        """
        Close and cleanup the Index.
        """

    @property
    @abstractmethod
    def index_id(self) -> str:
        """
        :return: Unique ID for this index
                 (e.g. same database/dataset storage + same index driver implementation = same id)
        """

    @abstractmethod
    def transaction(self) -> AbstractTransaction:
        """
        :return: a Transaction context manager for this index.
        """

    def thread_transaction(self) -> AbstractTransaction | None:
        """
        :return: The existing Transaction object cached in thread-local storage for this index, if there is one.
        """
        return thread_local_cache(f"txn-{self.index_id}", None)

    def __enter__(self) -> "AbstractIndex":
        return self

    def __exit__(self) -> None:
        self.close()


class AbstractIndexDriver(ABC):
    """
    Abstract base class for an IndexDriver.  All IndexDrivers should inherit from this base class
    and implement all abstract methods.
    """

    @classmethod
    @abstractmethod
    def index_class(cls) -> type[AbstractIndex]: ...

    @classmethod
    def connect_to_index(
        cls,
        config_env: ODCEnvironment,
        application_name: str | None = None,
        validate_connection: bool = True,
    ) -> "AbstractIndex":
        return cls.index_class().from_config(
            config_env, application_name, validate_connection
        )

    @staticmethod
    @abstractmethod
    @deprecat(
        reason="The 'metadata_type_from_doc' static method has been deprecated. "
        "Please use the 'index.metadata_type.from_doc()' instead.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def metadata_type_from_doc(definition: dict) -> MetadataType: ...

    @staticmethod
    def get_config_option_handlers(env: ODCEnvironment) -> Iterable[ODCOptionHandler]:
        """
        Default Implementation does nothing.
        Override for driver-specific config handling (e.g. for db connection)
        """
        return []



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest


def test_cli_product_subcommand(index_empty, clirunner, dataset_add_configs) -> None:
    runner = clirunner(["product", "update"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [FILES]" in runner.output
    assert "Update existing products." in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["product", "update", dataset_add_configs.empty_file],
        verbose_flag=False,
        expect_success=False,
    )
    assert "All files are empty, exit" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["product", "add"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [FILES]" in runner.output
    assert "Add or update products in" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["product", "list"], verbose_flag=False)
    assert "Usage:  [OPTIONS] [FILES]" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["product", "show", "ga_ls8c_ard_3"], verbose_flag=False, expect_success=False
    )
    assert "No products" not in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["product", "add", dataset_add_configs.empty_file],
        verbose_flag=False,
        expect_success=False,
    )
    assert "All files are empty, exit" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["product", "delete"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [PRODUCT_NAMES]" in runner.output
    assert "Delete products" in runner.output

    runner = clirunner(
        ["product", "delete", "ga_ls8c_ard_3"], verbose_flag=False, expect_success=False
    )
    assert '"ga_ls8c_ard_3" is not a valid Product name' in runner.output
    assert runner.exit_code == 1


def test_cli_metadata_subcommand(index_empty, clirunner, dataset_add_configs) -> None:
    runner = clirunner(["metadata", "update"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [FILES]" in runner.output
    assert "Update existing metadata types." in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["metadata", "update", dataset_add_configs.empty_file],
        verbose_flag=False,
        expect_success=False,
    )
    assert "All files are empty, exit" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["metadata", "add"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [FILES]" in runner.output
    assert "Add or update metadata types in" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["metadata", "add", dataset_add_configs.empty_file],
        verbose_flag=False,
        expect_success=False,
    )
    assert "All files are empty, exit" in runner.output
    assert runner.exit_code == 1


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_cli_dataset_subcommand(
    index,
    clirunner,
    extended_eo3_metadata_type,
    ls8_eo3_product,
    wo_eo3_product,
    africa_s2_eo3_product,
    eo3_dataset_paths,
) -> None:
    # Tests with no datasets in db
    runner = clirunner(["dataset", "add"], verbose_flag=False, expect_success=False)
    assert (
        "Indexing datasets  [####################################]  100%"
        not in runner.output
    )
    assert "Usage:  [OPTIONS] [DATASET_PATHS]" in runner.output
    assert "Add datasets" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["dataset", "update"], verbose_flag=False, expect_success=False)
    assert "0 successful, 0 failed" not in runner.output
    assert "Usage:  [OPTIONS] [DATASET_PATHS]" in runner.output
    assert "Update datasets" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["dataset", "info"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [IDS]" in runner.output
    assert "Display dataset information" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["dataset", "uri-search"], verbose_flag=False, expect_success=False
    )
    assert "Usage:  [OPTIONS] [PATHS]" in runner.output
    assert "Search by dataset locations" in runner.output
    assert runner.exit_code == 1

    # Insert datasets
    for path in eo3_dataset_paths:
        clirunner(["dataset", "add", "--ignore-lineage", path])

    runner = clirunner(
        ["dataset", "find-duplicates"], verbose_flag=False, expect_success=False
    )
    assert "Error: must provide field names to match on" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["dataset", "find-duplicates", "region_code", "fake_field"],
        verbose_flag=False,
        expect_success=False,
    )
    assert (
        "Error: no products found with fields region_code, fake_field" in runner.output
    )
    assert runner.exit_code == 1

    runner = clirunner(
        [
            "dataset",
            "find-duplicates",
            "region_code",
            "landsat_scene_id",
            "-p",
            "ga_ls8c_ard_3",
            "-p",
            "ga_ls_wo_3",
        ],
        verbose_flag=False,
        expect_success=False,
    )
    assert (
        "Error: specified products ga_ls_wo_3 do not contain all required fields"
        in runner.output
    )
    assert runner.exit_code == 1

    runner = clirunner(
        ["dataset", "find-duplicates", "region_code", "uri"], verbose_flag=False
    )
    assert "No potential duplicates found." in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "find-duplicates", "region_code", "dataset_maturity"],
        verbose_flag=False,
    )
    assert "No potential duplicates found." not in runner.output
    assert "region_code: 090086\ndataset_maturity: final" in runner.output
    assert "region_code: '101077'\ndataset_maturity: final" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "count", "--count-only", "ga_ls8c_ard_3", "ga_ls_wo_3"],
        verbose_flag=False,
    )
    assert runner.output == "5\n"
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "count", "ga_ls8c_ard_3", "ga_ls_wo_3"], verbose_flag=False
    )
    assert "product: ga_ls8c_ard_3\ncount: 4" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "count", "ga_ls8c_ard_3", "--query", 'region_code="090086"'],
        verbose_flag=False,
    )
    assert "count: 2" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["dataset", "count", "--count-only", "--period", "1 month", "ga_ls8c_ard_3"],
        verbose_flag=False,
        expect_success=False,
    )
    assert (
        "Error: cannot return total count when requesting time slicing" in runner.output
    )
    assert runner.exit_code == 1

    runner = clirunner(
        [
            "dataset",
            "count",
            "--period",
            "1 year",
            "--query",
            "time in [2013-01-01, 2017-01-01]",
            "ga_ls8c_ard_3",
        ],
        verbose_flag=False,
    )
    assert "time: '2013-01-01'\ncount: 2" in runner.output
    assert runner.exit_code == 0

    clirunner(
        ["dataset", "archive", "c21648b1-a6fa-4de0-9dc3-9c445d8b295a"],
        verbose_flag=False,
    )
    runner = clirunner(["dataset", "count", "ga_ls8c_ard_3"], verbose_flag=False)
    assert "count: 3" in runner.output
    runner = clirunner(
        ["dataset", "count", "--status", "all", "ga_ls8c_ard_3"], verbose_flag=False
    )
    assert "count: 4" in runner.output
    runner = clirunner(
        ["dataset", "count", "--status", "archived", "ga_ls8c_ard_3"],
        verbose_flag=False,
    )
    assert "count: 1" in runner.output

    runner = clirunner(
        ["dataset", "search", "foo"], verbose_flag=False, expect_success=False
    )
    assert "Invalid expression" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["dataset", "search", "product=ga_ls8c_ard_3", 'region_code="090086"'],
        verbose_flag=False,
    )
    assert "id: 4a30d008-4e82-4d67-99af-28bc1629f766" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(["dataset", "archive"], verbose_flag=False, expect_success=False)
    assert "Completed dataset archival." not in runner.output
    assert "Usage:  [OPTIONS] [IDS]" in runner.output
    assert "Archive datasets" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["dataset", "archive", "--all"], verbose_flag=False)
    assert "Archiving dataset:" in runner.output
    assert "Completed dataset archival." in runner.output
    assert "Usage:  [OPTIONS] [IDS]" not in runner.output
    assert "Archive datasets" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(["dataset", "restore"], verbose_flag=False, expect_success=False)
    assert "Usage:  [OPTIONS] [IDS]" in runner.output
    assert "Restore datasets" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["dataset", "restore", "--all"], verbose_flag=False)
    assert "restoring" in runner.output
    assert "Usage:  [OPTIONS] [IDS]" not in runner.output
    assert "Restore datasets" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(["dataset", "purge"], verbose_flag=False, expect_success=False)
    assert "Completed dataset purge." not in runner.output
    assert "Usage:  [OPTIONS] [IDS]" in runner.output
    assert "Purge archived datasets" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["dataset", "purge", "--all"], verbose_flag=False)
    assert "Completed dataset purge." in runner.output
    assert "Usage:  [OPTIONS] [IDS]" not in runner.output
    assert runner.exit_code == 0


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_read_and_update_metadata_product_dataset_command(
    index,
    clirunner,
    ext_eo3_mdt_path,
    eo3_product_paths,
    eo3_dataset_paths,
    eo3_dataset_update_path,
) -> None:
    clirunner(["metadata", "add", ext_eo3_mdt_path])
    rerun_add = clirunner(["metadata", "add", ext_eo3_mdt_path])
    assert "WARNING Metadata Type" in rerun_add.output
    assert "is already in the database" in rerun_add.output

    update = clirunner(["metadata", "update", ext_eo3_mdt_path])
    assert "WARNING No changes detected for metadata type" in update.output

    for prod_path in eo3_product_paths:
        clirunner(["product", "add", prod_path])
        rerun_add = clirunner(["product", "add", prod_path])
        assert "WARNING Product" in rerun_add.output
        assert "is already in the database" in rerun_add.output

        update = clirunner(["product", "update", prod_path])
        assert "WARNING No changes detected for product" in update.output

    # Update before add
    for ds_path in eo3_dataset_paths:
        update = clirunner(["dataset", "update", ds_path])
        assert "No such dataset in the database" in update.output
        assert "Failure while processing" in update.output

        clirunner(["dataset", "add", "--ignore-lineage", ds_path])
        rerun_add = clirunner(["dataset", "add", "--ignore-lineage", ds_path])
        assert "WARNING Dataset" in rerun_add.output
        assert "is already in the database" in rerun_add.output

    update = clirunner(["dataset", "update", eo3_dataset_update_path])
    assert "Unsafe changes in" in update.output
    assert "0 successful, 1 failed" in update.output

    update = clirunner(
        [
            "dataset",
            "update",
            "--allow-any",
            "properties.datetime",
            eo3_dataset_update_path,
        ]
    )
    assert "1 successful, 0 failed" in update.output



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Mapping
from typing import Any

from datacube.model.fields import Field, SimpleField
from datacube.model.fields import get_dataset_fields as generic_get_dataset_fields
from datacube.utils.changes import Offset


# TODO: SimpleFields cannot handle non-metadata fields because e.g. the extract API expects a doc, not a Dataset model
def get_native_fields() -> dict[str, Field]:
    return {
        "id": SimpleField(["id"], str, "string", name="id", description="Dataset UUID"),
        "product": SimpleField(
            ["product", "name"],
            str,
            "string",
            name="product",
            description="Product name",
        ),
        "label": SimpleField(["label"], str, "string", name="label", description=""),
        "format": SimpleField(
            ["format", "name"],
            str,
            "string",
            name="format",
            description="File format (GeoTIFF, NetCDF)",
        ),
        "metadata_doc": SimpleField(
            [], str, "string", name="metadata_doc", description="Full metadata document"
        ),
    }


def get_dataset_fields(metadata_definition: Mapping[str, Any]) -> dict[str, Field]:
    fields = get_native_fields()
    fields.update(generic_get_dataset_fields(metadata_definition))
    return fields


def build_custom_fields(custom_offsets: Mapping[str, Offset]) -> dict[str, Field]:
    return {
        name: SimpleField(
            list(offset),
            lambda x: x,
            "any",
            name=name,
            description=f"Custom field: {name}",
        )
        for name, offset in custom_offsets.items()
    }



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

from typing_extensions import override

from datacube.drivers.postgis import PostGisDb
from datacube.index.abstract import AbstractUserResource
from datacube.index.postgis._transaction import IndexResourceAddIn

if TYPE_CHECKING:
    from datacube.index.postgis.index import Index


class UserResource(AbstractUserResource, IndexResourceAddIn):
    def __init__(self, db: PostGisDb, index: Index) -> None:
        self._db = db
        self._index = index

    @override
    def grant_role(self, role: str, *usernames: str) -> None:
        """
        Grant a role to users
        """
        with self._db_connection() as connection:
            connection.grant_role(role, usernames)

    @override
    def create_user(
        self, username: str, password: str, role: str, description: str | None = None
    ) -> None:
        """
        Create a new user.
        """
        with self._db_connection() as connection:
            connection.create_user(username, password, role, description=description)

    @override
    def delete_user(self, *usernames: str) -> None:
        """
        Delete a user
        """
        with self._db_connection() as connection:
            connection.drop_users(usernames)

    @override
    def list_users(self) -> Iterable[tuple[str, str, str | None]]:
        """
        :return: list of (role, user, description)
        """
        with self._db_connection() as connection:
            yield from connection.list_users()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from collections.abc import Iterable, Iterator, Sequence
from contextlib import contextmanager
from pathlib import Path

from deprecat import deprecat
from odc.geo import CRS
from typing_extensions import override

from datacube.cfg.api import ODCEnvironment, ODCOptionHandler
from datacube.cfg.opt import config_options_for_psql_driver
from datacube.drivers.postgis import PostGisDb, PostgisDbAPI
from datacube.index.abstract import (
    DSID,
    AbstractIndex,
    AbstractIndexDriver,
    AbstractTransaction,
    default_metadata_type_docs,
)
from datacube.index.postgis._datasets import DatasetResource
from datacube.index.postgis._lineage import LineageResource
from datacube.index.postgis._metadata_types import MetadataTypeResource
from datacube.index.postgis._products import ProductResource
from datacube.index.postgis._transaction import PostgisTransaction
from datacube.index.postgis._users import UserResource
from datacube.migration import ODC2DeprecationWarning
from datacube.model import MetadataType

_LOG: logging.Logger = logging.getLogger(__name__)


_DEFAULT_METADATA_TYPES_PATH: Path = Path(__file__).parent.joinpath(
    "default-metadata-types.yaml"
)


class Index(AbstractIndex):
    """
    Access to the datacube index.

    DON'T INITIALISE THIS DIRECTLY (it will break in the future). Use `datacube.index.index_connect()` or
    access property ``.index`` on your existing :class:`datacube.api.core.Datacube`.

    These are thread safe. But not multiprocess safe once a connection is made (db connections cannot be shared
    between processes) You can close idle connections before forking by calling close(), provided you know no
    other connections are active. Or else use a separate instance of this class in each process.

    :ivar datacube.index._datasets.DatasetResource datasets: store and retrieve :class:`datacube.model.Dataset`
    :ivar datacube.index._products.ProductResource products: store and retrieve :class:`datacube.model.Product`\
    (formerly called DatasetType)
    :ivar datacube.index._metadata_types.MetadataTypeResource metadata_types: store and retrieve \
    :class:`datacube.model.MetadataType`
    :ivar UserResource users: user management
    """

    #   Metadata type support flags
    supports_eo3 = True

    #   Database/storage feature support flags
    supports_write = True
    supports_persistance = True
    supports_transactions = True
    supports_spatial_indexes = True

    #   User management support flags
    supports_users = True

    #   Lineage support flags
    supports_lineage = True
    supports_external_lineage = True
    supports_external_home = True

    def __init__(self, db: PostGisDb, env: ODCEnvironment) -> None:
        self._db = db
        self._env = env

        self._users = UserResource(db, self)
        self._metadata_types = MetadataTypeResource(db, self)
        self._products = ProductResource(db, self)
        self._lineage = LineageResource(db, self)
        self._datasets = DatasetResource(db, self)

    @property
    @override
    def name(self) -> str:
        return "pgis_index"

    @property
    @override
    def environment(self) -> ODCEnvironment:
        return self._env

    @property
    @override
    def users(self) -> UserResource:
        return self._users

    @property
    @override
    def metadata_types(self) -> MetadataTypeResource:
        return self._metadata_types

    @property
    @override
    def products(self) -> ProductResource:
        return self._products

    @property
    @override
    def lineage(self) -> LineageResource:
        return self._lineage

    @property
    @override
    def datasets(self) -> DatasetResource:
        return self._datasets

    @property
    @override
    def url(self) -> str:
        return str(self._db.url)

    @classmethod
    @override
    def from_config(
        cls,
        config_env: ODCEnvironment,
        application_name: str | None = None,
        validate_connection: bool = True,
    ) -> "Index":
        db = PostGisDb.from_config(
            config_env,
            application_name=application_name,
            validate_connection=validate_connection,
        )
        return cls(db, config_env)

    @classmethod
    @override
    def get_dataset_fields(cls, doc) -> dict:
        return PostGisDb.get_dataset_fields(doc)

    @override
    def init_db(
        self,
        with_default_types: bool = True,
        with_permissions: bool = True,
        with_default_spatial_index: bool = True,
    ) -> bool:
        is_new = self._db.init(with_permissions=with_permissions)

        if is_new and with_default_types:
            _LOG.info("Adding default metadata types.")
            for doc in default_metadata_type_docs(_DEFAULT_METADATA_TYPES_PATH):
                self.metadata_types.add(
                    self.metadata_types.from_doc(doc), allow_table_lock=True
                )

        if is_new and with_default_spatial_index:
            self.create_spatial_index(CRS("EPSG:4326"))

        return is_new

    @override
    def close(self) -> None:
        """
        Close any idle connections database connections.

        This is good practice if you are keeping the Index instance in scope
        but won't be using it for a while.

        (Connections are normally closed automatically when this object is deleted: ie. no references exist)
        """
        self._db.close()

    @property
    @override
    def index_id(self) -> str:
        return self.url

    @override
    def transaction(self) -> AbstractTransaction:
        return PostgisTransaction(self._db, self.index_id)

    @override
    def create_spatial_index(self, crs: CRS) -> bool:
        sp_idx = self._db.create_spatial_index(crs)
        return sp_idx is not None

    @override
    def spatial_indexes(self, refresh: bool = False) -> Iterable[CRS]:
        return self._db.spatially_indexed_crses(refresh)

    @override
    def update_spatial_index(
        self,
        crses: Sequence[CRS] = [],
        product_names: Sequence[str] = [],
        dataset_ids: Sequence[DSID] = [],
    ) -> int:
        with self._active_connection(transaction=True) as conn:
            return conn.update_spindex(crses, product_names, dataset_ids)

    @override
    def __repr__(self) -> str:
        return f"Index<db={self._db!r}>"

    @override
    def drop_spatial_index(self, crs: CRS) -> bool:
        return self._db.drop_spatial_index(crs)

    @contextmanager
    def _active_connection(self, transaction: bool = False) -> Iterator[PostgisDbAPI]:
        """
        Context manager representing a database connection.

        If there is an active transaction for this index in the current thread, the connection object from that
        transaction is returned, with the active transaction remaining in control of commit and rollback.

        If there is no active transaction and the transaction argument is True, a new transactionised connection
        is returned, with this context manager handling commit and rollback.

        If there is no active transaction and the transaction argument is False (the default), a new connection
        is returned with autocommit semantics.

        Note that autocommit behaviour is NOT available if there is an active transaction for the index
        and the active thread.

        :param transaction: Use a transaction if one is not already active for the thread.
        :return: A PostgresDbAPI object, with the specified transaction semantics.
        """
        trans = self.thread_transaction()
        if trans is not None:
            # Use active transaction
            yield trans._connection
        elif transaction:
            with self._db._connect() as conn:
                conn.begin()
                # assert conn.in_transaction
                try:
                    yield conn
                    conn.commit()
                except Exception:  # pylint: disable=broad-except
                    conn.rollback()
                    raise
        else:
            # Autocommit behaviour:
            with self._db._connect() as conn:
                yield conn


class PostgisIndexDriver(AbstractIndexDriver):
    @classmethod
    @override
    def index_class(cls) -> type[AbstractIndex]:
        return Index

    @staticmethod
    @override
    @deprecat(
        reason="The 'metadata_type_from_doc' static method has been deprecated. "
        "Please use the 'index.metadata_type.from_doc()' instead.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def metadata_type_from_doc(definition: dict) -> MetadataType:
        """
        :param definition:
        """
        # TODO: Validate metadata is ODCv2 compliant - e.g. either non-raster or EO3.
        MetadataType.validate(definition)  # type: ignore
        return MetadataType(definition, search_field_extractor=Index.get_dataset_fields)

    @staticmethod
    @override
    def get_config_option_handlers(env: ODCEnvironment) -> Iterable[ODCOptionHandler]:
        return config_options_for_psql_driver(env)


def index_driver_init() -> PostgisIndexDriver:
    return PostgisIndexDriver()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Datacube
========

Provides access to multi-dimensional data, with a focus on Earth observations data such as LANDSAT.

To use this module, see the `Developer Guide <https://opendatacube.readthedocs.io/en/latest/installation/index.html>`_.

The main class to access the datacube is :class:`datacube.Datacube`.

"""

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "Unknown/Not Installed"

from .api import Datacube
from .utils import xarray_geoextensions

__all__ = [
    "Datacube",
    "__version__",
    "xarray_geoextensions",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Modules for creating and accessing Data Store Units

"""

from ..drivers.datasource import DataSource, GeoRasterReader, RasterShape, RasterWindow
from ._base import BandInfo, measurement_paths
from ._load import reproject_and_fuse

__all__ = [
    "BandInfo",
    "DataSource",
    "GeoRasterReader",
    "RasterShape",
    "RasterWindow",
    "measurement_paths",
    "reproject_and_fuse",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

from datacube.model import Dataset, Product


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_crs_parse(indexed_ls5_scene_products: list[Product]) -> None:
    product = indexed_ls5_scene_products[2]

    # Explicit CRS, should load fine.
    # Taken from LS8_OLI_NBAR_3577_-14_-11_20140601021126000000.nc
    d = Dataset(
        product,
        {
            "grid_spatial": {
                "projection": {
                    "valid_data": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                [-1396453.986271351, -1100000.0],
                                [-1400000.0, -1100000.0],
                                [-1400000.0, -1053643.4714392645],
                                [-1392296.4215373022, -1054399.795365491],
                                [-1390986.9858215596, -1054531.808155645],
                                [-1390806.366757733, -1054585.3982497198],
                                [-1396453.986271351, -1100000.0],
                            ]
                        ],
                    },
                    "geo_ref_points": {
                        "ll": {"x": -1400000.0, "y": -1100000.0},
                        "lr": {"x": -1300000.0, "y": -1100000.0},
                        "ul": {"x": -1400000.0, "y": -1000000.0},
                        "ur": {"x": -1300000.0, "y": -1000000.0},
                    },
                    "spatial_reference": "EPSG:3577",
                }
            }
        },
    )
    assert str(d.crs) == "EPSG:3577"
    assert d.extent is not None

    # No projection specified in the dataset
    ds = Dataset(product, {})
    assert ds.crs is None
    assert ds.extent is None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Iterable

from typing_extensions import override

from datacube.index.abstract import AbstractMetadataTypeResource
from datacube.model import MetadataType
from datacube.utils.changes import Change
from datacube.utils.documents import JsonDict


class MetadataTypeResource(AbstractMetadataTypeResource):
    @override
    def from_doc(self, definition: JsonDict) -> MetadataType:
        raise NotImplementedError

    @override
    def add(
        self, metadata_type: MetadataType, allow_table_lock: bool = False
    ) -> MetadataType:
        raise NotImplementedError

    @override
    def can_update(
        self, metadata_type: MetadataType, allow_unsafe_updates: bool = False
    ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        raise NotImplementedError

    @override
    def update(
        self,
        metadata_type: MetadataType,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> MetadataType:
        raise NotImplementedError

    @override
    def get_unsafe(self, id_: int) -> MetadataType:
        raise KeyError(id_)

    @override
    def get_by_name_unsafe(self, name: str) -> MetadataType:
        raise KeyError(name)

    @override
    def check_field_indexes(
        self,
        allow_table_lock: bool = False,
        rebuild_views: bool = False,
        rebuild_indexes: bool = False,
    ) -> None:
        raise NotImplementedError

    @override
    def get_all(self) -> Iterable[MetadataType]:
        return []



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Iterable, MutableMapping
from typing import Literal

from alembic import context
from alembic.operations import MigrationScript
from alembic.runtime.migration import MigrationContext

from datacube.cfg import ODCConfig, ODCEnvironment
from datacube.drivers.postgis._connections import PostGisDb
from datacube.drivers.postgis._schema import Base
from datacube.drivers.postgis._spatial import is_spindex_table_name
from datacube.drivers.postgis.sql import SCHEMA_NAME

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
try:
    config = context.config
except AttributeError:
    # Occurs when being scanned for doctests, so safe to ignore type warnings
    config = None  # type: ignore[assignment]

# Interpret the config file for Python logging.
# This line sets up loggers basically.
#
# Doesn't play nice with rest of ODC's management of logging.
#
#  if config.config_file_name is not None:
#     fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.
    """
    context.configure(
        dialect_name="postgresql",
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        version_table_schema="odc",
    )

    with context.begin_transaction():
        context.run_migrations()


def include_name(
    name,
    type_: Literal[
        "schema",
        "table",
        "column",
        "index",
        "unique_constraint",
        "foreign_key_constraint",
    ],
    parent_names: MutableMapping[
        Literal["schema_name", "table_name", "schema_qualified_table_name"], str | None
    ],
) -> bool:
    if type_ == "table":
        # Ignore postgis system table
        if name == "spatial_ref_sys" and parent_names["schema_name"] is None:
            return False

        # Ignore dynamically generated spatial index tables
        return not is_spindex_table_name(name)
    elif type_ == "schema":
        # Monitor default and odc schema
        return name is None or name == SCHEMA_NAME
    elif type_ == "column":
        # Include all columns
        return True
    else:
        # Include any constraints, indexes, etc, that made it this far.
        return True


def get_odc_env() -> ODCEnvironment:
    # In active Alembic Config?
    cfg = config.attributes.get("cfg")
    env = config.attributes.get("env")
    raw_config = config.attributes.get("raw_config")
    if not (cfg or env or raw_config):
        # No?  How about from alembic CLI -X args?
        x_args = context.get_x_argument(as_dictionary=True)
        cfg = x_args.get("cfg")
        if cfg:
            cfg = cfg.split(":")
        env = x_args.get("env")
        raw_config = x_args.get("raw_config")
    return ODCConfig.get_environment(env=env, config=cfg, raw_config=raw_config)


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """
    # An active postgis Connection:
    connection = config.attributes.get("connection")
    if connection:
        run_migration_with_connection(connection)
        return

    # ODC index:
    index = config.attributes.get("index")
    if index:
        connectable = index._db._engine
    else:
        db = PostGisDb.create(
            get_odc_env(), application_name="migration", validate=False
        )
        connectable = db._engine

    with connectable.connect() as connection:
        run_migration_with_connection(connection)


def run_migration_with_connection(connection) -> None:
    # Do not generate a file with --autogenerate unless there is a difference.
    def process_revision_directives(
        context: MigrationContext,
        revision: str | Iterable[str | None] | Iterable[str],
        directives: list[MigrationScript],
    ) -> None:
        assert config.cmd_opts is not None
        if getattr(config.cmd_opts, "autogenerate", False):
            script = directives[0]
            assert script.upgrade_ops is not None
            if script.upgrade_ops.is_empty():
                directives[:] = []

    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        process_revision_directives=process_revision_directives,
        version_table_schema=SCHEMA_NAME,
        include_schemas=True,
        include_name=include_name,
    )
    with context.begin_transaction():
        context.run_migrations()


if config is None:
    # See comment above
    pass
elif context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Dask Distributed Tools"""

import logging
import os
import queue
import threading
from collections.abc import Iterable
from random import randint
from typing import Any

import dask
import toolz
from dask.delayed import Delayed, DelayedLeaf, delayed
from dask.distributed import Client

__all__ = [
    "compute_tasks",
    "partition_map",
    "pmap",
    "save_blob_to_file",
    "save_blob_to_s3",
    "start_local_dask",
]

_LOG: logging.Logger = logging.getLogger(__name__)


def get_total_available_memory(check_jupyter_hub: bool = True) -> int:
    """Figure out how much memory is available
    1. Check MEM_LIMIT environment variable, set by jupyterhub
    2. Use hardware information if that not set
    """
    if check_jupyter_hub:
        mem_limit = os.environ.get("MEM_LIMIT", None)
        if mem_limit is not None:
            return int(mem_limit)

    from psutil import virtual_memory

    return virtual_memory().total


def compute_memory_per_worker(
    n_workers: int = 1,
    mem_safety_margin: str | int | None = None,
    memory_limit: str | int | None = None,
) -> int:
    """Figure out how much memory to assign per worker.

    result can be passed into ``memory_limit=`` parameter of dask worker/cluster/client
    """
    from dask.utils import parse_bytes

    if isinstance(memory_limit, str):
        memory_limit = parse_bytes(memory_limit)

    if isinstance(mem_safety_margin, str):
        mem_safety_margin = parse_bytes(mem_safety_margin)

    if memory_limit is None and mem_safety_margin is None:
        total_bytes = get_total_available_memory()
        # leave 500Mb or half of all memory if RAM is less than 1 Gb
        mem_safety_margin = min(500 * (1024 * 1024), total_bytes // 2)
    elif memory_limit is None:
        total_bytes = get_total_available_memory()
    elif mem_safety_margin is None:
        total_bytes = memory_limit
        mem_safety_margin = 0
    else:
        total_bytes = memory_limit
    assert mem_safety_margin is not None  # For typechecker.

    return (total_bytes - mem_safety_margin) // n_workers


def start_local_dask(
    n_workers: int = 1,
    threads_per_worker: int | None = None,
    mem_safety_margin: str | int | None = None,
    memory_limit: str | int | None = None,
    **kw,
):
    """
    Wrapper around ``distributed.Client(..)`` constructor that deals with memory better.

    It also configures ``distributed.dashboard.link`` to go over proxy when operating
    from behind jupyterhub.

    :param n_workers: number of worker processes to launch
    :param threads_per_worker: number of threads per worker, default is as many as there are CPUs
    :param memory_limit: maximum memory to use across all workers
    :param mem_safety_margin: bytes to reserve for the rest of the system, only applicable
                              if ``memory_limit=`` is not supplied.

    .. note::

        if ``memory_limit=`` is supplied, it will be parsed and divided equally between workers.
    """
    # if dashboard.link set to default value and running behind hub, make dashboard link go via proxy
    if (
        dask.config.get("distributed.dashboard.link")
        == "{scheme}://{host}:{port}/status"
    ):
        jup_prefix = os.environ.get("JUPYTERHUB_SERVICE_PREFIX")
        if jup_prefix is not None:
            jup_prefix = jup_prefix.rstrip("/")
            dask.config.set(
                {"distributed.dashboard.link": f"{jup_prefix}/proxy/{{port}}/status"}
            )

    memory_limit = compute_memory_per_worker(
        n_workers=n_workers,
        memory_limit=memory_limit,
        mem_safety_margin=mem_safety_margin,
    )

    client = Client(
        n_workers=n_workers,
        threads_per_worker=threads_per_worker,
        memory_limit=memory_limit,
        **kw,
    )

    return client


def _randomize(prefix: str) -> str:
    return f"{prefix}-{randint(0, 0xFFFFFFFF):08x}"


def partition_map(
    n: int, func: Any, its: Iterable[Any], name: str = "compute"
) -> Iterable[Any]:
    """
    Parallel map in lumps.

    Partition sequence into lumps of size ``n``, then construct dask delayed computation evaluating to:

    .. code-block:: python

        [func(x) for x in its[0:1n]],
        [func(x) for x in its[n:2n]],
        ...
        [func(x) for x in its[..]],

    This is useful when you need to process a large number of small (quick) tasks (pixel drill for example).

    :param n: number of elements to process in one go
    :param func: Function to apply (non-dask)
    :param its:  Values to feed to fun
    :param name: How the computation should be named in dask visualizations

    Returns
    -------

    Iterator of ``dask.Delayed`` objects.
    """

    def lump_proc(dd):
        return [func(d) for d in dd]

    proc = delayed(lump_proc, nout=1, pure=True)
    data_name = _randomize("data_" + name)
    name = _randomize(name)

    for i, dd in enumerate(toolz.partition_all(n, its)):
        lump = delayed(dd, pure=True, traverse=False, name=data_name + str(i))
        yield proc(lump, dask_key_name=name + str(i))


def compute_tasks(
    tasks: Iterable[Any], client: Client, max_in_flight: int = 3
) -> Iterable[Any]:
    """Parallel compute stream with back pressure.

    Equivalent to:


    .. code-block:: python

        (client.compute(task).result()
          for task in tasks)

    but with up to ``max_in_flight`` tasks being processed at the same time.
    Input/Output order is preserved, so there is a possibility of head of
    line blocking.

    .. note::

          lower limit is 3 concurrent tasks to simplify implementation,
          there is no point calling this function if you want one active
          task and supporting exactly 2 active tasks is not worth the complexity,
          for now. We might special-case 2 at some point.
    """
    # New thread:
    #    1. Take dask task from iterator
    #    2. Submit to client for processing
    #    3. Send it of to wrk_q
    #
    # Calling thread:
    #    1. Pull scheduled future from wrk_q
    #    2. Wait for result of the future
    #    3. yield result to calling code
    from .generic import it2q, qmap

    # (max_in_flight - 2) -- one on each side of queue
    wrk_q: queue.Queue = queue.Queue(maxsize=max(1, max_in_flight - 2))

    # fifo_timeout='0ms' ensures that priority of later tasks is lower
    futures = (client.compute(task, fifo_timeout="0ms") for task in tasks)

    in_thread = threading.Thread(target=it2q, args=(futures, wrk_q))
    in_thread.start()

    yield from qmap(lambda f: f.result(), wrk_q)

    in_thread.join()


def pmap(
    func: Any,
    its: Iterable[Any],
    client: Client,
    lump: int = 1,
    max_in_flight: int = 3,
    name: str = "compute",
) -> Iterable[Any]:
    """Parallel map with back pressure.

    Equivalent to this:

    .. code-block:: python

       (func(x) for x in its)

    Except that ``func(x)`` runs concurrently on dask cluster.

    :param func:   Method that will be applied concurrently to data from ``its``
    :param its:    Iterator of input values
    :param client: Connected dask client
    :param lump:   Group this many datasets into one task
    :param max_in_flight: Maximum number of active tasks to submit
    :param name:   Dask name for computation
    """
    max_in_flight = max_in_flight // lump

    tasks = partition_map(lump, func, its, name=name)

    for xx in compute_tasks(tasks, client=client, max_in_flight=max_in_flight):
        yield from xx


def _save_blob_to_file(
    data: bytes | str, fname: str, with_deps=None
) -> tuple[str, bool]:
    if isinstance(data, str):
        data = data.encode("utf8")

    try:
        with open(fname, "wb") as f:
            f.write(data)
    except OSError:
        return fname, False

    return fname, True


def _save_blob_to_s3(
    data: bytes | str,
    url: str,
    profile: str | None = None,
    creds=None,
    region_name: str | None = None,
    with_deps=None,
    **kw,
) -> tuple[str, bool]:
    from botocore.errorfactory import ClientError
    from botocore.exceptions import BotoCoreError

    from .aws import s3_client, s3_dump

    try:
        s3 = s3_client(
            profile=profile, creds=creds, region_name=region_name, cache=True
        )

        result = s3_dump(data, url, s3=s3, **kw)
    except (OSError, BotoCoreError, ClientError):
        result = False

    return url, result


_save_blob_to_file_delayed = delayed(
    _save_blob_to_file, name="save-to-disk", pure=False
)
_save_blob_to_s3_delayed = delayed(_save_blob_to_s3, name="save-to-s3", pure=False)


def save_blob_to_file(
    data: bytes | str, fname: str, with_deps=None
) -> Delayed | DelayedLeaf:
    """
    Dump from memory to local filesystem as a dask delayed operation.

    :param data: Data blob to save to file (have to fit into memory all at once),
                 strings will be saved in UTF8 format.

    :param fname: Path to file

    :param with_deps: Useful for introducing dependencies into dask graph,
                      for example save yaml file after saving all tiff files.

    Returns
    -------
    ``(FilePath, True)`` tuple on success
    ``(FilePath, False)`` on any error


    .. note::

       Dask workers must be local or have network filesystem mounted in
       the same path as calling code.
    """
    return _save_blob_to_file_delayed(data, fname, with_deps=with_deps)


def save_blob_to_s3(
    data: bytes | str,
    url: str,
    profile: str | None = None,
    creds=None,
    region_name: str | None = None,
    with_deps=None,
    **kw,
) -> Delayed | DelayedLeaf:
    """
    Dump from memory to S3 as a dask delayed operation.

    :param data:        Data blob to save to file (have to fit into memory all at once)
    :param url:         Url in a form s3://bucket/path/to/file
    :param profile:     Profile name to lookup (only used if session is not supplied)
    :param creds:       Override credentials with supplied data
    :param region_name: Region name to use, overrides session setting

    :param with_deps:   Useful for introducing dependencies into dask graph,
                        for example save yaml file after saving all tiff files.
    :param kw:          Passed on to ``s3.put_object(..)``, useful for things like ContentType/ACL

    Returns
    -------

    ``(url, True)`` tuple on success
    ``(url, False)`` on any error
    """
    return _save_blob_to_s3_delayed(
        data,
        url,
        profile=profile,
        creds=creds,
        region_name=region_name,
        with_deps=with_deps,
        **kw,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Useful functions for Datacube users

Not used internally, those should go in `utils.py`
"""

import numpy as np

DEFAULT_PROFILE = {
    "blockxsize": 256,
    "blockysize": 256,
    "compress": "lzw",
    "driver": "GTiff",
    "interleave": "band",
    "nodata": 0.0,
    "tiled": True,
}


def _calculate_blocksize(profile) -> None:
    # Block size must be smaller than the image size, and for geotiffs must be divisible by 16
    # Fix for small images.
    if profile["blockxsize"] > profile["width"]:
        if profile["width"] % 16 == 0 or profile["width"] < 16:
            profile["blockxsize"] = profile["width"]
        else:
            profile["blockxsize"] = 16

    if profile["blockysize"] > profile["height"]:
        if profile["height"] % 16 == 0 or profile["height"] < 16:
            profile["blockysize"] = profile["height"]
        else:
            profile["blockysize"] = 16


def ga_pq_fuser(dest, src) -> None:
    """
    Fuse two Geoscience Australia Pixel Quality ndarrays

    To be used as a `fuse_func` when loaded `grouped` data, for example when grouping
    by solar day to avoid duplicate data from scene overlaps.
    """
    valid_bit = 8
    valid_val = 1 << valid_bit

    no_data_dest_mask = ~(dest & valid_val).astype(bool)
    np.copyto(dest, src, where=no_data_dest_mask)

    both_data_mask = (valid_val & dest & src).astype(bool)
    np.copyto(dest, src & dest, where=both_data_mask)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from collections.abc import Mapping
from typing import Any

from deprecat import deprecat
from odc.geo import CRS
from typing_extensions import override

from datacube.cfg import ODCEnvironment
from datacube.index.abstract import (
    AbstractIndex,
    AbstractIndexDriver,
    NoLineageResource,
    UnhandledTransaction,
)
from datacube.index.null._datasets import DatasetResource
from datacube.index.null._metadata_types import MetadataTypeResource
from datacube.index.null._products import ProductResource
from datacube.index.null._users import UserResource
from datacube.migration import ODC2DeprecationWarning
from datacube.model import Field, MetadataType
from datacube.model.fields import get_dataset_fields

_LOG: logging.Logger = logging.getLogger(__name__)


class Index(AbstractIndex):
    """
    (Sub-)Minimal (non-)implementation of the Index API.
    """

    #   Metadata type support flags
    supports_legacy = True
    supports_eo3 = True
    supports_nongeo = True

    #   User management support flags
    supports_users = True

    def __init__(self, env: ODCEnvironment) -> None:
        self._env = env
        self._users = UserResource()
        self._metadata_types = MetadataTypeResource()
        self._products = ProductResource(self)
        self._lineage = NoLineageResource(self)
        self._datasets = DatasetResource(self)

    @property
    @override
    def name(self) -> str:
        return "null_index"

    @property
    @override
    def environment(self) -> ODCEnvironment:
        return self._env

    @property
    @override
    def users(self) -> UserResource:
        return self._users

    @property
    @override
    def metadata_types(self) -> MetadataTypeResource:
        return self._metadata_types

    @property
    @override
    def products(self) -> ProductResource:
        return self._products

    @property
    @override
    def lineage(self) -> NoLineageResource:
        return self._lineage

    @property
    @override
    def datasets(self) -> DatasetResource:
        return self._datasets

    @property
    @override
    def url(self) -> str:
        return "null"

    @property
    @override
    def index_id(self) -> str:
        return "null"

    @override
    def transaction(self) -> UnhandledTransaction:
        return UnhandledTransaction(self.index_id)

    @classmethod
    @override
    def from_config(
        cls,
        cfg_env: ODCEnvironment,
        application_name: str | None = None,
        validate_connection: bool = True,
    ) -> "Index":
        return cls(cfg_env)

    @classmethod
    @override
    def get_dataset_fields(cls, doc: Mapping[str, Any]) -> dict[str, Field]:
        return get_dataset_fields(doc)

    @override
    def init_db(
        self, with_default_types: bool = True, with_permissions: bool = True
    ) -> bool:
        return True

    @override
    def close(self) -> None:
        pass

    @override
    def create_spatial_index(self, crs: CRS) -> bool:
        _LOG.warning("null driver does not support spatio-temporal indexes")
        return False

    @override
    def __repr__(self) -> str:
        return "Index<null>"


class NullIndexDriver(AbstractIndexDriver):
    @classmethod
    @override
    def index_class(cls) -> type[AbstractIndex]:
        return Index

    @staticmethod
    @override
    @deprecat(
        reason="The 'metadata_type_from_doc' static method has been deprecated. "
        "Please use the 'index.metadata_type.from_doc()' instead.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def metadata_type_from_doc(definition: dict) -> MetadataType:
        """
        :param definition:
        """
        MetadataType.validate(definition)  # type: ignore
        return MetadataType(definition, search_field_extractor=Index.get_dataset_fields)


def index_driver_init() -> NullIndexDriver:
    return NullIndexDriver()
